(##declare
  (standard-bindings)
  (extended-bindings)
  (block)
  (fixnum)
  (not safe)
)

;;; Functional and unstable

(define (vector-copy v)
  (let* ((length (vector-length v))
	 (result (make-vector length)))
    (let loop ((n 0))
      (vector-set! result n (vector-ref v n))
      (if (= n length)
	  v
	  (loop (+ n 1))))))

(define (sort obj pred)
  (define (loop l)
    (if (and (pair? l) (pair? (cdr l)))
	(split l '() '())
	l))

  (define (split l one two)
    (if (pair? l)
	(split (cdr l) two (cons (car l) one))
	(merge (loop one) (loop two))))

  (define (merge one two)
    (cond ((null? one) two)
	  ((pred (car two) (car one))
	   (cons (car two)
		 (merge (cdr two) one)))
	  (else
	   (cons (car one)
		 (merge (cdr one) two)))))

  (cond ((or (pair? obj) (null? obj))
	 (loop obj))
	((vector? obj)
	 (sort! (vector-copy obj) pred))
	(else
	 (error "sort: argument should be a list or vector" obj))))

;; This merge sort is stable for partial orders (for predicates like
;; <=, rather than like <).

(define (sort! v pred)
  (define (sort-internal! vec temp low high)
    (if (< low high)
	(let* ((middle (quotient (+ low high) 2))
	       (next (+ middle 1)))
	  (sort-internal! temp vec low middle)
	  (sort-internal! temp vec next high)
	  (let loop ((p low) (p1 low) (p2 next))
	    (if (not (> p high))
		(cond ((> p1 middle)
		       (vector-set! vec p (vector-ref temp p2))
		       (loop (+ p 1) p1 (+ p2 1)))
		      ((or (> p2 high)
			   (pred (vector-ref temp p1)
				 (vector-ref temp p2)))
		       (vector-set! vec p (vector-ref temp p1))
		       (loop (+ p 1) (+ p1 1) p2))
		      (else
		       (vector-set! vec p (vector-ref temp p2))
		       (loop (+ p 1) p1 (+ p2 1)))))))))

  (if (not (vector? v))
      (error "sort!: argument not a vector" v))

  (sort-internal! v
		  (vector-copy v)
		  0
		  (- (vector-length v) 1))
  v)

;; SET OPERATIONS
; (representation as lists with distinct elements)

(define (adjoin element set)
  (if (memq element set) set (cons element set)))

(define (eliminate element set)
  (cond ((null? set) set)
	((eq? element (car set)) (cdr set))
	(else (cons (car set) (eliminate element (cdr set))))))

(define (intersect list1 list2)
  (let loop ((l list1))
    (cond ((null? l) '())
	  ((memq (car l) list2) (cons (car l) (loop (cdr l))))
	  (else (loop (cdr l))))))

(define (union list1 list2)
  (if (null? list1)
      list2
      (union (cdr list1)
	     (adjoin (car list1) list2))))

;; GRAPH NODES

; (define-structure
;   (internal-node
;    (print-procedure (unparser/standard-method
; 		     'graph-node
; 		     (lambda (state node)
; 		       (unparse-object state (internal-node-name node))))))
;   name
;   (green-edges '())
;   (red-edges '())
;   blue-edges)

; Above is MIT version; below is portable

(define make-internal-node vector)
(define (internal-node-name node) (vector-ref node 0))
(define (internal-node-green-edges node) (vector-ref node 1))
(define (internal-node-red-edges node) (vector-ref node 2))
(define (internal-node-blue-edges node) (vector-ref node 3))
(define (set-internal-node-name! node name) (vector-set! node 0 name))
(define (set-internal-node-green-edges! node edges) (vector-set! node 1 edges))
(define (set-internal-node-red-edges! node edges) (vector-set! node 2 edges))
(define (set-internal-node-blue-edges! node edges) (vector-set! node 3 edges))

; End of portability stuff

(define (make-node name . blue-edges)	; User's constructor
  (let ((name (if (symbol? name) (symbol->string name) name))
	(blue-edges (if (null? blue-edges) 'not-a-node-yet (car blue-edges))))
    (make-internal-node name '() '() blue-edges)))

(define (copy-node node)
  (make-internal-node (name node) '() '() (blue-edges node)))

; Selectors

(define name internal-node-name)
(define (make-edge-getter selector)
  (lambda (node)
    (if (or (none-node? node) (any-node? node))
	(error "Can't get edges from the ANY or NONE nodes")
	(selector node))))
(define red-edges (make-edge-getter internal-node-red-edges))
(define green-edges (make-edge-getter internal-node-green-edges))
(define blue-edges (make-edge-getter internal-node-blue-edges))

; Mutators

(define (make-edge-setter mutator!)
  (lambda (node value)
    (cond ((any-node? node) (error "Can't set edges from the ANY node"))
	  ((none-node? node) 'ok)
	  (else (mutator! node value)))))
(define set-red-edges! (make-edge-setter set-internal-node-red-edges!))
(define set-green-edges! (make-edge-setter set-internal-node-green-edges!))
(define set-blue-edges! (make-edge-setter set-internal-node-blue-edges!))

;; BLUE EDGES

; (define-structure
;   (blue-edge
;    (print-procedure
;     (unparser/standard-method
;      'blue-edge
;      (lambda (state edge)
;        (unparse-object state (blue-edge-operation edge))))))
;   operation arg-node res-node)

; Above is MIT version; below is portable

(define make-blue-edge vector)
(define (blue-edge-operation edge) (vector-ref edge 0))
(define (blue-edge-arg-node edge) (vector-ref edge 1))
(define (blue-edge-res-node edge) (vector-ref edge 2))
(define (set-blue-edge-operation! edge value) (vector-set! edge 0 value))
(define (set-blue-edge-arg-node! edge value) (vector-set! edge 1 value))
(define (set-blue-edge-res-node! edge value) (vector-set! edge 2 value))

; End of portability stuff

; Selectors
(define operation blue-edge-operation)
(define arg-node blue-edge-arg-node)
(define res-node blue-edge-res-node)

; Mutators
(define set-arg-node! set-blue-edge-arg-node!)
(define set-res-node! set-blue-edge-res-node!)

; Higher level operations on blue edges

(define (lookup-op op node)
  (let loop ((edges (blue-edges node)))
    (cond ((null? edges) '())
	  ((eq? op (operation (car edges))) (car edges))
	  (else (loop (cdr edges))))))

(define (has-op? op node)
  (not (null? (lookup-op op node))))

; Add a (new) blue edge to a node

; (define (adjoin-blue-edge! blue-edge node)
;   (let ((current-one (lookup-op (operation blue-edge) node)))
;     (cond ((null? current-one)
; 	       (set-blue-edges! node
; 		 (cons blue-edge (blue-edges node))))
; 	  ((and (eq? (arg-node current-one) (arg-node blue-edge))
; 		(eq? (res-node current-one) (res-node blue-edge)))
; 	   'ok)
; 	   (else (error "Two non-equivalent blue edges for op"
; 			blue-edge node)))))

;; GRAPHS

; (define-structure
;   (internal-graph
;    (print-procedure
;     (unparser/standard-method 'graph
;      (lambda (state edge)
;        (unparse-object state (map name (internal-graph-nodes edge)))))))
;   nodes already-met already-joined)

; Above is MIT version; below is portable

(define make-internal-graph vector)
(define (internal-graph-nodes graph) (vector-ref graph 0))
(define (internal-graph-already-met graph) (vector-ref graph 1))
(define (internal-graph-already-joined graph) (vector-ref graph 2))
(define (set-internal-graph-nodes! graph nodes) (vector-set! graph 0 nodes))

; End of portability stuff

; Constructor

(define (make-graph . nodes)
  (make-internal-graph nodes (make-empty-table) (make-empty-table)))

; Selectors

(define graph-nodes internal-graph-nodes)
(define already-met internal-graph-already-met)
(define already-joined internal-graph-already-joined)

; Higher level functions on graphs

(define (add-graph-nodes! graph nodes)
  (set-internal-graph-nodes! graph (cons nodes (graph-nodes graph))))

(define (copy-graph g)
  (define (copy-list l) (vector->list (list->vector l)))
  (make-internal-graph
   (copy-list (graph-nodes g))
   (already-met g)
   (already-joined g)))

(define (clean-graph g)
  (define (clean-node node)
    (if (not (or (any-node? node) (none-node? node)))
	(begin
	  (set-green-edges! node '())
	  (set-red-edges! node '()))))
  (for-each clean-node (graph-nodes g))
  g)

(define (canonicalize-graph graph classes)
  (define (fix node)
    (define (fix-set object selector mutator)
      (mutator object 
	       (map (lambda (node)
		      (find-canonical-representative node classes))
		    (selector object))))
    (if (not (or (none-node? node) (any-node? node)))
	(begin
	  (fix-set node green-edges set-green-edges!)
	  (fix-set node red-edges set-red-edges!)
	  (for-each 
	   (lambda (blue-edge)
	     (set-arg-node! blue-edge
			    (find-canonical-representative (arg-node blue-edge) classes))
	     (set-res-node! blue-edge
			    (find-canonical-representative (res-node blue-edge) classes)))
	   (blue-edges node))))
    node)
  (define (fix-table table)
    (define (canonical? node) (eq? node (find-canonical-representative node classes)))
    (define (filter-and-fix predicate-fn update-fn list)
      (let loop ((list list))
	(cond ((null? list) '())
	      ((predicate-fn (car list))
	       (cons (update-fn (car list)) (loop (cdr list))))
	      (else (loop (cdr list))))))
    (define (fix-line line)
      (filter-and-fix
       (lambda (entry) (canonical? (car entry)))
       (lambda (entry) (cons (car entry)
			     (find-canonical-representative (cdr entry) classes)))
       line))
    (if (null? table)
	'()
	(cons (car table)
	      (filter-and-fix
	       (lambda (entry) (canonical? (car entry)))
	       (lambda (entry) (cons (car entry) (fix-line (cdr entry))))
	       (cdr table)))))
  (make-internal-graph
   (map (lambda (class) (fix (car class))) classes)
   (fix-table (already-met graph))
   (fix-table (already-joined graph))))

;; USEFUL NODES

(define none-node (make-node 'none '#t))
(define (none-node? node) (eq? node none-node))

(define any-node (make-node 'any '()))
(define (any-node? node) (eq? node any-node))

;; COLORED EDGE TESTS

(define (green-edge? from-node to-node)
  (cond ((any-node? from-node) '#f)
	((none-node? from-node) '#t)
	((memq to-node (green-edges from-node)) '#t)
	(else '#f)))

(define (red-edge? from-node to-node)
  (cond ((any-node? from-node) '#f)
	((none-node? from-node) '#t)
	((memq to-node (red-edges from-node)) '#t)
	(else '#f)))

;; SIGNATURE

; Return signature (i.e. <arg, res>) given an operation and a node

(define sig
  (let ((none-comma-any (cons none-node any-node)))
    (lambda (op node)			; Returns (arg, res)
      (let ((the-edge (lookup-op op node)))
	(if (not (null? the-edge))
	    (cons (arg-node the-edge) (res-node the-edge))
	    none-comma-any)))))

; Selectors from signature

(define (arg pair) (car pair))
(define (res pair) (cdr pair))

;; CONFORMITY

(define (conforms? t1 t2)
  (define nodes-with-red-edges-out '())
  (define (add-red-edge! from-node to-node)
    (set-red-edges! from-node (adjoin to-node (red-edges from-node)))
    (set! nodes-with-red-edges-out
	  (adjoin from-node nodes-with-red-edges-out)))
  (define (greenify-red-edges! from-node)
    (set-green-edges! from-node
		      (append (red-edges from-node) (green-edges from-node)))
    (set-red-edges! from-node '()))
  (define (delete-red-edges! from-node)
    (set-red-edges! from-node '()))
  (define (does-conform t1 t2)
    (cond ((or (none-node? t1) (any-node? t2)) '#t)
	  ((or (any-node? t1) (none-node? t2)) '#f)
	  ((green-edge? t1 t2) '#t)
	  ((red-edge? t1 t2) '#t)
	  (else
	   (add-red-edge! t1 t2)
	   (let loop ((blues (blue-edges t2)))
	     (if (null? blues)
		 '#t
		 (let* ((current-edge (car blues))
			(phi (operation current-edge)))
		   (and (has-op? phi t1)
			(does-conform
			 (res (sig phi t1))
			 (res (sig phi t2)))
			(does-conform
			 (arg (sig phi t2))
			 (arg (sig phi t1)))
			(loop (cdr blues)))))))))
  (let ((result (does-conform t1 t2)))
    (for-each (if result greenify-red-edges! delete-red-edges!)
	      nodes-with-red-edges-out)
    result))

(define (equivalent? a b)
  (and (conforms? a b) (conforms? b a)))

;; EQUIVALENCE CLASSIFICATION
; Given a list of nodes, return a list of equivalence classes

(define (classify nodes)
  (let node-loop ((classes '())
		  (nodes nodes))
    (if (null? nodes)
	(map (lambda (class)
	       (sort class
		     (lambda (node1 node2)
		       (< (string-length (name node1))
			  (string-length (name node2))))))
	     classes)
	(let ((this-node (car nodes)))
	  (define (add-node classes)
	    (cond ((null? classes) (list (list this-node)))
		  ((equivalent? this-node (caar classes))
		   (cons (cons this-node (car classes))
			 (cdr classes)))
		  (else (cons (car classes)
			      (add-node (cdr classes))))))
	  (node-loop (add-node classes)
		     (cdr nodes))))))

; Given a node N and a classified set of nodes,
; find the canonical member corresponding to N

(define (find-canonical-representative element classification)
  (let loop ((classes classification))
    (cond ((null? classes) (error "Can't classify" element)) 
	  ((memq element (car classes)) (car (car classes)))
	  (else (loop (cdr classes))))))

; Reduce a graph by taking only one member of each equivalence 
; class and canonicalizing all outbound pointers

(define (reduce graph)
  (let ((classes (classify (graph-nodes graph))))
    (canonicalize-graph graph classes)))

;; TWO DIMENSIONAL TABLES

(define (make-empty-table) (list 'table))
(define (lookup table x y)
  (let ((one (assq x (cdr table))))
    (if one
	(let ((two (assq y (cdr one))))
	  (if two (cdr two) '#f))
	'#f)))
(define (insert! table x y value)
  (define (make-singleton-table x y)
    (list (cons x y)))
  (let ((one (assq x (cdr table))))
    (if one
	(set-cdr! one (cons (cons y value) (cdr one)))
	(set-cdr! table (cons (cons x (make-singleton-table y value))
			      (cdr table))))))

;; MEET/JOIN 
; These update the graph when computing the node for node1*node2

(define (blue-edge-operate arg-fn res-fn graph op sig1 sig2)
  (make-blue-edge op
		  (arg-fn graph (arg sig1) (arg sig2))
		  (res-fn graph (res sig1) (res sig2))))

(define (meet graph node1 node2)
  (cond ((eq? node1 node2) node1)
	((or (any-node? node1) (any-node? node2)) any-node) ; canonicalize
	((none-node? node1) node2)
	((none-node? node2) node1)
	((lookup (already-met graph) node1 node2)) ; return it if found
	((conforms? node1 node2) node2)
	((conforms? node2 node1) node1)
	(else
	 (let ((result
		(make-node (string-append "(" (name node1) " ^ " (name node2) ")"))))
	   (add-graph-nodes! graph result)
	   (insert! (already-met graph) node1 node2 result)
	   (set-blue-edges! result
	     (map
	      (lambda (op)
		(blue-edge-operate join meet graph op (sig op node1) (sig op node2)))
	      (intersect (map operation (blue-edges node1))
			 (map operation (blue-edges node2)))))
	   result))))

(define (join graph node1 node2)
  (cond ((eq? node1 node2) node1)
	((any-node? node1) node2)
	((any-node? node2) node1)
	((or (none-node? node1) (none-node? node2)) none-node) ; canonicalize
	((lookup (already-joined graph) node1 node2)) ; return it if found
	((conforms? node1 node2) node1)
	((conforms? node2 node1) node2)
	(else
	 (let ((result
		(make-node (string-append "(" (name node1) " v " (name node2) ")"))))
	   (add-graph-nodes! graph result)
	   (insert! (already-joined graph) node1 node2 result)
	   (set-blue-edges! result
             (map
	      (lambda (op)
		(blue-edge-operate meet join graph op (sig op node1) (sig op node2)))
	      (union (map operation (blue-edges node1))
		     (map operation (blue-edges node2)))))
	   result))))

;; MAKE A LATTICE FROM A GRAPH

(define (make-lattice g print?)
  (define (step g)
    (let* ((copy (copy-graph g))
	   (nodes (graph-nodes copy)))
      (for-each (lambda (first)
		  (for-each (lambda (second)
			      (meet copy first second) (join copy first second))
			    nodes))
		nodes)
      copy))
  (define (loop g count)
    (if print? (display count))
    (let ((lattice (step g)))
      (if print? (begin (display " -> ") (display (length (graph-nodes lattice)))))
      (let* ((new-g (reduce lattice))
	     (new-count (length (graph-nodes new-g))))
	(if (= new-count count)
	    (begin
	      (if print? (newline))
	      new-g)
	    (begin
	      (if print? (begin (display " -> ") (display new-count) (newline)))
	      (loop new-g new-count))))))
  (let ((graph
	 (apply make-graph
		(adjoin any-node (adjoin none-node (graph-nodes (clean-graph g)))))))
    (loop graph (length (graph-nodes graph)))))

;; DEBUG and TEST

(define a '())
(define b '())
(define c '())
(define d '())

(define (reset)
  (set! a (make-node 'a))
  (set! b (make-node 'b))
  (set-blue-edges! a (list (make-blue-edge 'phi any-node b)))
  (set-blue-edges! b (list (make-blue-edge 'phi any-node a)
			   (make-blue-edge 'theta any-node b)))
  (set! c (make-node "c"))
  (set! d (make-node "d"))
  (set-blue-edges! c (list (make-blue-edge 'theta any-node b)))
  (set-blue-edges! d (list (make-blue-edge 'phi any-node c)
			   (make-blue-edge 'theta any-node d)))
  '(made a b c d))

(define (go)
  (reset)
  (map name
       (graph-nodes (make-lattice (make-graph a b c d any-node none-node) '#f))))

;;; call: (go)

(define (show msg proc)
  (newline);  (display msg) (pp proc)
)

(define (try thunk)
  (call-with-current-continuation
    (lambda (cont)
      (##catch-all (lambda (s args)
                     (display "-----> ")
                     (write s)
                     (show " on " thunk)
                     (cont 0))
                   (lambda ()
                     (write (thunk))
                     (show " on " thunk))))))

(##define-macro (test form expect)
  `(test-form (lambda () ,form) ',expect))

(define (test-form thunk expect)
  (if (equal? (thunk) expect)
    (write 'ok)
    (write (list 'expected expect)))
  (newline))

(##define-macro (err form)
  `(let ()
     (##declare (safe))
     (err-form (lambda () ,form))))

(define (err-form thunk)
  (test-form
    (lambda ()
      (call-with-current-continuation
        (lambda (abort)
          (##catch-all
            (lambda lst
              (write lst)
              (newline)
              (abort 'cookie))
            thunk))))
    'cookie))

(define (f1) 'ok)
(define (f2 a) (list a))
(define (f3 . a) (list a))
(define (f4 a . b) (list a b))

(define (f5 a #!optional) (list a))
(define (f6 a #!optional b) (list a b))
(define (f7 a #!optional (b (list a b))) (list a b))

(define (f8 a #!rest b) (list a b))

(define (f9 a #!key) (list a))
(define (f10 a #!key b) (list a b))
(define (f11 a #!key (b (list a b))) (list a b))

(define (f12 a #!optional #!rest b) (list a b))
(define (f13 a #!optional b #!rest c) (list a b c))

(define (f14 a #!optional #!key) (list a))
(define (f15 a #!optional #!key b) (list a b))
(define (f16 a #!optional #!key (b (list a b))) (list a b))
(define (f17 a #!optional b #!key) (list a b))
(define (f18 a #!optional b #!key c) (list a b c))
(define (f19 a #!optional (b (list a b c)) #!key (c (list a b c))) (list a b c))

(define (f20 a #!rest b #!key) (list a b))
(define (f21 a #!rest b #!key c) (list a b c))
(define (f22 a #!rest b #!key (c (list a b c))) (list a b c))

(define (f23 a #!optional #!rest b #!key) (list a b))
(define (f24 a #!optional #!rest b #!key c) (list a b c))
(define (f25 a #!optional #!rest b #!key (c (list a b c))) (list a b c))
(define (f26 a #!optional b #!rest c #!key) (list a b c))
(define (f27 a #!optional (b (list a b c)) #!rest c #!key) (list a b c))
(define (f28 a #!optional b #!rest c #!key d) (list a b c d))
(define (f29 a #!optional (b (list a b c d)) #!rest c #!key (d (list a b c d))) (list a b c d))

;(define a 6)
;(define b 7)
;(define c 8)
;(define d 9)

(define (start f)
  (newline)
  (show "" f))

(define (run)

  (set! a 6)
  (set! b 7)
  (set! c 8)
  (set! d 9)

  (start f1)
  (try (lambda () (f1)))
  (try (lambda () (f1 1)))
  (try (lambda () (f1 1 2)))
  (try (lambda () (f1 1 2 3)))
  (try (lambda () (f1 1 2 3 4)))
  (try (lambda () (f1 x: 1 y: 2)))

  (start f2)
  (try (lambda () (f2)))
  (try (lambda () (f2 1)))
  (try (lambda () (f2 1 2)))
  (try (lambda () (f2 1 2 3)))
  (try (lambda () (f2 1 2 3 4)))
  (try (lambda () (f2 1 x: 2 y: 3)))

  (start f3)
  (try (lambda () (f3)))
  (try (lambda () (f3 1)))
  (try (lambda () (f3 1 2)))
  (try (lambda () (f3 1 2 3)))
  (try (lambda () (f3 1 2 3 4)))
  (try (lambda () (f3 x: 1 y: 2)))

  (start f4)
  (try (lambda () (f4)))
  (try (lambda () (f4 1)))
  (try (lambda () (f4 1 2)))
  (try (lambda () (f4 1 2 3)))
  (try (lambda () (f4 1 2 3 4)))
  (try (lambda () (f4 1 x: 2 y: 3)))

  (start f5)
  (try (lambda () (f5)))
  (try (lambda () (f5 1)))
  (try (lambda () (f5 1 2)))
  (try (lambda () (f5 1 2 3)))
  (try (lambda () (f5 1 2 3 4)))
  (try (lambda () (f5 1 x: 2 y: 3)))

  (start f6)
  (try (lambda () (f6)))
  (try (lambda () (f6 1)))
  (try (lambda () (f6 1 2)))
  (try (lambda () (f6 1 2 3)))
  (try (lambda () (f6 1 2 3 4)))
  (try (lambda () (f6 1 2 x: 3 y: 4)))

  (start f7)
  (try (lambda () (f7)))
  (try (lambda () (f7 1)))
  (try (lambda () (f7 1 2)))
  (try (lambda () (f7 1 2 3)))
  (try (lambda () (f7 1 2 3 4)))
  (try (lambda () (f7 1 2 x: 3 y: 4)))

  (start f8)
  (try (lambda () (f8)))
  (try (lambda () (f8 1)))
  (try (lambda () (f8 1 2)))
  (try (lambda () (f8 1 2 3)))
  (try (lambda () (f8 1 2 3 4)))
  (try (lambda () (f8 1 x: 2 y: 3)))

  (start f9)
  (try (lambda () (f9)))
  (try (lambda () (f9 1)))
  (try (lambda () (f9 1 2)))
  (try (lambda () (f9 1 2 3)))
  (try (lambda () (f9 1 x: 2 y: 3)))

  (start f10)
  (try (lambda () (f10)))
  (try (lambda () (f10 1)))
  (try (lambda () (f10 1 2)))
  (try (lambda () (f10 1 2 3)))
  (try (lambda () (f10 1 b: 2)))
  (try (lambda () (f10 1 x: 2)))
  (try (lambda () (f10 1 x: 2 b: 3)))

  (start f11)
  (try (lambda () (f11)))
  (try (lambda () (f11 1)))
  (try (lambda () (f11 1 2)))
  (try (lambda () (f11 1 2 3)))
  (try (lambda () (f11 1 b: 2)))
  (try (lambda () (f11 1 x: 2)))
  (try (lambda () (f11 1 x: 2 b: 3)))

  (start f12)
  (try (lambda () (f12)))
  (try (lambda () (f12 1)))
  (try (lambda () (f12 1 2)))
  (try (lambda () (f12 1 2 3)))
  (try (lambda () (f12 1 x: 2 y: 3)))

  (start f13)
  (try (lambda () (f13)))
  (try (lambda () (f13 1)))
  (try (lambda () (f13 1 2)))
  (try (lambda () (f13 1 2 3)))
  (try (lambda () (f13 1 2 3 4)))
  (try (lambda () (f13 1 2 x: 3 y: 4)))

  (start f14)
  (try (lambda () (f14)))
  (try (lambda () (f14 1)))
  (try (lambda () (f14 1 2)))
  (try (lambda () (f14 1 2 3)))
  (try (lambda () (f14 1 x: 2)))
  (try (lambda () (f14 1 x: 2 y: 3)))

  (start f15)
  (try (lambda () (f15)))
  (try (lambda () (f15 1)))
  (try (lambda () (f15 1 2)))
  (try (lambda () (f15 1 2 3)))
  (try (lambda () (f15 1 b: 2)))
  (try (lambda () (f15 1 x: 2)))
  (try (lambda () (f15 1 x: 2 b: 3)))

  (start f16)
  (try (lambda () (f16)))
  (try (lambda () (f16 1)))
  (try (lambda () (f16 1 2)))
  (try (lambda () (f16 1 2 3)))
  (try (lambda () (f16 1 b: 2)))
  (try (lambda () (f16 1 x: 2)))
  (try (lambda () (f16 1 x: 2 b: 3)))

  (start f17)
  (try (lambda () (f17)))
  (try (lambda () (f17 1)))
  (try (lambda () (f17 1 2)))
  (try (lambda () (f17 1 2 3)))
  (try (lambda () (f17 1 2 3 4)))
  (try (lambda () (f17 1 2 x: 3)))
  (try (lambda () (f17 1 2 x: 3 y: 4)))

  (start f18)
  (try (lambda () (f18)))
  (try (lambda () (f18 1)))
  (try (lambda () (f18 1 2)))
  (try (lambda () (f18 1 2 3)))
  (try (lambda () (f18 1 2 3 4)))
  (try (lambda () (f18 1 2 c: 3)))
  (try (lambda () (f18 1 2 x: 3)))
  (try (lambda () (f18 1 2 x: 3 c: 4)))

  (start f19)
  (try (lambda () (f19)))
  (try (lambda () (f19 1)))
  (try (lambda () (f19 1 2)))
  (try (lambda () (f19 1 2 3)))
  (try (lambda () (f19 1 2 3 4)))
  (try (lambda () (f19 1 2 c: 3)))
  (try (lambda () (f19 1 2 x: 3)))
  (try (lambda () (f19 1 2 x: 3 c: 4)))

  (start f20)
  (try (lambda () (f20)))
  (try (lambda () (f20 1)))
  (try (lambda () (f20 1 2)))
  (try (lambda () (f20 1 2 3)))
  (try (lambda () (f20 1 x: 2 y: 3)))

  (start f21)
  (try (lambda () (f21)))
  (try (lambda () (f21 1)))
  (try (lambda () (f21 1 2)))
  (try (lambda () (f21 1 2 3)))
  (try (lambda () (f21 1 c: 2)))
  (try (lambda () (f21 1 x: 2)))
  (try (lambda () (f21 1 x: 2 c: 3)))

  (start f22)
  (try (lambda () (f22)))
  (try (lambda () (f22 1)))
  (try (lambda () (f22 1 2)))
  (try (lambda () (f22 1 2 3)))
  (try (lambda () (f22 1 c: 2)))
  (try (lambda () (f22 1 x: 2)))
  (try (lambda () (f22 1 x: 2 c: 3)))

  (start f23)
  (try (lambda () (f23)))
  (try (lambda () (f23 1)))
  (try (lambda () (f23 1 2)))
  (try (lambda () (f23 1 2 3)))
  (try (lambda () (f23 1 x: 2 y: 3)))

  (start f24)
  (try (lambda () (f24)))
  (try (lambda () (f24 1)))
  (try (lambda () (f24 1 2)))
  (try (lambda () (f24 1 2 3)))
  (try (lambda () (f24 1 c: 2)))
  (try (lambda () (f24 1 x: 2)))
  (try (lambda () (f24 1 x: 2 c: 3)))

  (start f25)
  (try (lambda () (f25)))
  (try (lambda () (f25 1)))
  (try (lambda () (f25 1 2)))
  (try (lambda () (f25 1 2 3)))
  (try (lambda () (f25 1 c: 2)))
  (try (lambda () (f25 1 x: 2)))
  (try (lambda () (f25 1 x: 2 c: 3)))

  (start f26)
  (try (lambda () (f26)))
  (try (lambda () (f26 1)))
  (try (lambda () (f26 1 2)))
  (try (lambda () (f26 1 2 3)))
  (try (lambda () (f26 1 2 3 4)))
  (try (lambda () (f26 1 2 x: 3 y: 4)))

  (start f27)
  (try (lambda () (f27)))
  (try (lambda () (f27 1)))
  (try (lambda () (f27 1 2)))
  (try (lambda () (f27 1 2 3)))
  (try (lambda () (f27 1 2 3 4)))
  (try (lambda () (f27 1 2 c: 3)))
  (try (lambda () (f27 1 2 x: 3)))
  (try (lambda () (f27 1 2 x: 3 c: 4)))

  (start f28)
  (try (lambda () (f28)))
  (try (lambda () (f28 1)))
  (try (lambda () (f28 1 2)))
  (try (lambda () (f28 1 2 3)))
  (try (lambda () (f28 1 2 3 4)))
  (try (lambda () (f28 1 2 c: 3)))
  (try (lambda () (f28 1 2 x: 3)))
  (try (lambda () (f28 1 2 x: 3 c: 4)))

  (start f29)
  (try (lambda () (f29)))
  (try (lambda () (f29 1)))
  (try (lambda () (f29 1 2)))
  (try (lambda () (f29 1 2 3)))
  (try (lambda () (f29 1 2 3 4)))
  (try (lambda () (f29 1 2 c: 3)))
  (try (lambda () (f29 1 2 x: 3)))
  (try (lambda () (f29 1 2 x: 3 c: 4)))

  (newline)

  (test (string? "5678") #t)
  (test (string? 123456789012345678) #f)
  (test (make-string 0) "")
  (test (make-string 3) "   ")
  (test (make-string 5 #\6) "66666")
  (test (string) "")
  (test (string #\5 #\6) "56")
  (test (string-length "5678") 4)
  (test (string-ref "5678" 3) #\8)
  (test (let ((x (string #\5 #\6))) (string-set! x 1 #\3) x) "53")
  (test (string->list "56") (#\5 #\6))
  (test (list->string '(#\5 #\6)) "56")
  (err (make-string -1))
  (err (make-string 8388607))
  (err (make-string 8388608))
  (err (make-string 536870911))
  (err (make-string 536870912))
  (err (make-string 1.5))
  (err (make-string 1 'a))
  (err (string #\5 'b))
  (err (string-length 123456789012345678))
  (err (string-ref 123456789012345678 0))
  (err (string-ref "56" -1))
  (err (string-ref "56" 2))
  (err (string-ref "56" 536870912))
  (err (string-set! 123456789012345678 0 #\3))
  (err (let ((x (string #\5 #\6))) (string-set! x -1 #\3) x))
  (err (let ((x (string #\5 #\6))) (string-set! x 2 #\3) x))
  (err (let ((x (string #\5 #\6))) (string-set! x 536870912 #\3) x))
  (err (let ((x (string #\5 #\6))) (string-set! x 1 'a) x))
  (err (string->list 123456789012345678))
  (err (list->string 123456789012345678))
  (err (list->string '(#\5 b)))

  (test (vector? '#(5 6 7 8)) #t)
  (test (vector? 123456789012345678) #f)
  (test (make-vector 0) #())
  (test (make-vector 3) #(#f #f #f))
  (test (make-vector 5 'a) #(a a a a a))
  (test (vector) #())
  (test (vector 5 'b) #(5 b))
  (test (vector-length '#(5 6 7 8)) 4)
  (test (vector-ref '#(5 6 7 8) 3) 8)
  (test (let ((x (vector 5 6))) (vector-set! x 1 3) x) #(5 3))
  (test (vector->list '#(5 6)) (5 6))
  (test (list->vector '(5 b)) #(5 b))
  (err (make-vector -1))
  (err (make-vector 4194303))
  (err (make-vector 4194304))
  (err (make-vector 536870911))
  (err (make-vector 536870912))
  (err (make-vector 1.5))
  (err (vector-length 123456789012345678))
  (err (vector-ref 123456789012345678 0))
  (err (vector-ref '#(5 6) -1))
  (err (vector-ref '#(5 6) 2))
  (err (vector-ref '#(5 6) 536870912))
  (err (vector-set! 123456789012345678 0 3))
  (err (let ((x (vector 5 6))) (vector-set! x -1 3) x))
  (err (let ((x (vector 5 6))) (vector-set! x 2 3) x))
  (err (let ((x (vector 5 6))) (vector-set! x 536870912 3) x))
  (err (vector->list 123456789012345678))
  (err (list->vector 123456789012345678))

  (test (u8vector? '#u8(5 6 7 8)) #t)
  (test (u8vector? 123456789012345678) #f)
  (test (make-u8vector 0) #u8())
  (test (make-u8vector 3) #u8(0 0 0))
  (test (make-u8vector 5 6) #u8(6 6 6 6 6))
  (test (u8vector) #u8())
  (test (u8vector 5 -1 -128 255) #u8(5 255 128 255))
  (test (u8vector-length '#u8(5 6 7 8)) 4)
  (test (u8vector-ref '#u8(5 6 7 8) 3) 8)
  (test (let ((x (u8vector 5 6))) (u8vector-set! x 1 3) x) #u8(5 3))
  (test (let ((x (u8vector 5 6))) (u8vector-set! x 1 -128) x) #u8(5 128))
  (test (let ((x (u8vector 5 6))) (u8vector-set! x 1 255) x) #u8(5 255))
  (test (u8vector->list '#u8(5 6)) (5 6))
  (test (list->u8vector '(5 -1 -128 255)) #u8(5 255 128 255))
  (err (make-u8vector -1))
  (err (make-u8vector 16777215))
  (err (make-u8vector 16777216))
  (err (make-u8vector 536870911))
  (err (make-u8vector 536870912))
  (err (make-u8vector 1.5))
  (err (make-u8vector 1 'a))
  (err (u8vector 5 'b))
  (err (u8vector-length 123456789012345678))
  (err (u8vector-ref 123456789012345678 0))
  (err (u8vector-ref '#u8(5 6) -1))
  (err (u8vector-ref '#u8(5 6) 2))
  (err (u8vector-ref '#u8(5 6) 536870912))
  (err (u8vector-set! 123456789012345678 0 3))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x -1 3) x))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x 2 3) x))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x 536870912 3) x))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x 1 -129) x))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x 1 256) x))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x 1 536870912) x))
  (err (let ((x (u8vector 5 6))) (u8vector-set! x 1 'a) x))
  (err (u8vector->list 123456789012345678))
  (err (list->u8vector 123456789012345678))
  (err (list->u8vector '(5 b)))
  (err (list->u8vector '(256)))

  (test (u16vector? '#u16(5 6 7 8)) #t)
  (test (u16vector? 123456789012345678) #f)
  (test (make-u16vector 0) #u16())
  (test (make-u16vector 3) #u16(0 0 0))
  (test (make-u16vector 5 6) #u16(6 6 6 6 6))
  (test (u16vector) #u16())
  (test (u16vector 5 -1 -32768 65535) #u16(5 65535 32768 65535))
  (test (u16vector-length '#u16(5 6 7 8)) 4)
  (test (u16vector-ref '#u16(5 6 7 8) 3) 8)
  (test (let ((x (u16vector 5 6))) (u16vector-set! x 1 3) x) #u16(5 3))
  (test (let ((x (u16vector 5 6))) (u16vector-set! x 1 -32768) x) #u16(5 32768))
  (test (let ((x (u16vector 5 6))) (u16vector-set! x 1 65535) x) #u16(5 65535))
  (test (u16vector->list '#u16(5 6)) (5 6))
  (test (list->u16vector '(5 -1 -32768 65535)) #u16(5 65535 32768 65535))
  (err (make-u16vector -1))
  (err (make-u16vector 8388607))
  (err (make-u16vector 8388608))
  (err (make-u16vector 536870911))
  (err (make-u16vector 536870912))
  (err (make-u16vector 1.5))
  (err (make-u16vector 1 'a))
  (err (u16vector 5 'b))
  (err (u16vector-length 123456789012345678))
  (err (u16vector-ref 123456789012345678 0))
  (err (u16vector-ref '#u16(5 6) -1))
  (err (u16vector-ref '#u16(5 6) 2))
  (err (u16vector-ref '#u16(5 6) 536870912))
  (err (u16vector-set! 123456789012345678 0 3))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x -1 3) x))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x 2 3) x))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x 536870912 3) x))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x 1 -32769) x))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x 1 65536) x))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x 1 536870912) x))
  (err (let ((x (u16vector 5 6))) (u16vector-set! x 1 'a) x))
  (err (u16vector->list 123456789012345678))
  (err (list->u16vector 123456789012345678))
  (err (list->u16vector '(5 b)))
  (err (list->u16vector '(65536)))

  (test (u32vector? '#u32(5 6 7 8)) #t)
  (test (u32vector? 123456789012345678) #f)
  (test (make-u32vector 0) #u32())
  (test (make-u32vector 3) #u32(0 0 0))
  (test (make-u32vector 5 6) #u32(6 6 6 6 6))
  (test (u32vector) #u32())
  (test (u32vector 5 -1 -2147483648 4294967295) #u32(5 4294967295 2147483648 4294967295))
  (test (u32vector-length '#u32(5 6 7 8)) 4)
  (test (u32vector-ref '#u32(5 6 7 8) 3) 8)
  (test (let ((x (u32vector 5 6))) (u32vector-set! x 1 3) x) #u32(5 3))
  (test (let ((x (u32vector 5 6))) (u32vector-set! x 1 -2147483648) x) #u32(5 2147483648))
  (test (let ((x (u32vector 5 6))) (u32vector-set! x 1 4294967295) x) #u32(5 4294967295))
  (test (u32vector->list '#u32(5 6)) (5 6))
  (test (list->u32vector '(5 -1 -2147483648 4294967295)) #u32(5 4294967295 2147483648 4294967295))
  (err (make-u32vector -1))
  (err (make-u32vector 4194303))
  (err (make-u32vector 4194304))
  (err (make-u32vector 536870911))
  (err (make-u32vector 536870912))
  (err (make-u32vector 1.5))
  (err (make-u32vector 1 'a))
  (err (u32vector 5 'b))
  (err (u32vector-length 123456789012345678))
  (err (u32vector-ref 123456789012345678 0))
  (err (u32vector-ref '#u32(5 6) -1))
  (err (u32vector-ref '#u32(5 6) 2))
  (err (u32vector-ref '#u32(5 6) 536870912))
  (err (u32vector-set! 123456789012345678 0 3))
  (err (let ((x (u32vector 5 6))) (u32vector-set! x -1 3) x))
  (err (let ((x (u32vector 5 6))) (u32vector-set! x 2 3) x))
  (err (let ((x (u32vector 5 6))) (u32vector-set! x 536870912 3) x))
  (err (let ((x (u32vector 5 6))) (u32vector-set! x 1 -2147483649) x))
  (err (let ((x (u32vector 5 6))) (u32vector-set! x 1 4294967296) x))
  (err (let ((x (u32vector 5 6))) (u32vector-set! x 1 'a) x))
  (err (u32vector->list 123456789012345678))
  (err (list->u32vector 123456789012345678))
  (err (list->u32vector '(5 b)))
  (err (list->u32vector '(4294967296)))

  (test (f32vector? '#f32(5. 6. 7. 8.)) #t)
  (test (f32vector? 123456789012345678) #f)
  (test (make-f32vector 0) #f32())
  (test (make-f32vector 3) #f32(0. 0. 0.))
  (test (make-f32vector 5 6.) #f32(6. 6. 6. 6. 6.))
  (test (f32vector) #f32())
  (test (f32vector 5. 6.) #f32(5. 6.))
  (test (f32vector-length '#f32(5. 6. 7. 8.)) 4)
  (test (f32vector-ref '#f32(5. 6. 7. 8.) 3) 8.)
  (test (let ((x (f32vector 5. 6.))) (f32vector-set! x 1 3.) x) #f32(5. 3.))
  (test (f32vector->list '#f32(5. 6.)) (5. 6.))
  (test (list->f32vector '(5. 6.)) #f32(5. 6.))
  (err (make-f32vector -1))
  (err (make-f32vector 4194303))
  (err (make-f32vector 4194304))
  (err (make-f32vector 536870911))
  (err (make-f32vector 536870912))
  (err (make-f32vector 1.5))
  (err (make-f32vector 1 'a))
  (err (f32vector 5. 'b))
  (err (f32vector-length 123456789012345678))
  (err (f32vector-ref 123456789012345678 0))
  (err (f32vector-ref '#f32(5. 6.) -1))
  (err (f32vector-ref '#f32(5. 6.) 2))
  (err (f32vector-ref '#f32(5. 6.) 536870912))
  (err (f32vector-set! 123456789012345678 0 3.))
  (err (let ((x (f32vector 5. 6.))) (f32vector-set! x -1 3.) x))
  (err (let ((x (f32vector 5. 6.))) (f32vector-set! x 2 3.) x))
  (err (let ((x (f32vector 5. 6.))) (f32vector-set! x 536870912 3.) x))
  (err (let ((x (f32vector 5. 6.))) (f32vector-set! x 1 'a) x))
  (err (f32vector->list 123456789012345678))
  (err (list->f32vector 123456789012345678))
  (err (list->f32vector '(5. b)))

  (test (f64vector? '#f64(5. 6. 7. 8.)) #t)
  (test (f64vector? 123456789012345678) #f)
  (test (make-f64vector 0) #f64())
  (test (make-f64vector 3) #f64(0. 0. 0.))
  (test (make-f64vector 5 6.) #f64(6. 6. 6. 6. 6.))
  (test (f64vector) #f64())
  (test (f64vector 5. 6.) #f64(5. 6.))
  (test (f64vector-length '#f64(5. 6. 7. 8.)) 4)
  (test (f64vector-ref '#f64(5. 6. 7. 8.) 3) 8.)
  (test (let ((x (f64vector 5. 6.))) (f64vector-set! x 1 3.) x) #f64(5. 3.))
  (test (f64vector->list '#f64(5. 6.)) (5. 6.))
  (test (list->f64vector '(5. 6.)) #f64(5. 6.))
  (err (make-f64vector -1))
  (err (make-f64vector 2097151))
  (err (make-f64vector 2097152))
  (err (make-f64vector 536870911))
  (err (make-f64vector 536870912))
  (err (make-f64vector 1.5))
  (err (make-f64vector 1 'a))
  (err (f64vector 5. 'b))
  (err (f64vector-length 123456789012345678))
  (err (f64vector-ref 123456789012345678 0))
  (err (f64vector-ref '#f64(5. 6.) -1))
  (err (f64vector-ref '#f64(5. 6.) 2))
  (err (f64vector-ref '#f64(5. 6.) 536870912))
  (err (f64vector-set! 123456789012345678 0 3.))
  (err (let ((x (f64vector 5. 6.))) (f64vector-set! x -1 3.) x))
  (err (let ((x (f64vector 5. 6.))) (f64vector-set! x 2 3.) x))
  (err (let ((x (f64vector 5. 6.))) (f64vector-set! x 536870912 3.) x))
  (err (let ((x (f64vector 5. 6.))) (f64vector-set! x 1 'a) x))
  (err (f64vector->list 123456789012345678))
  (err (list->f64vector 123456789012345678))
  (err (list->f64vector '(5. b)))

)

(define intrs 0)
(##clear-timer-interrupt-jobs)
(##add-timer-interrupt-job (lambda () (set! intrs (+ intrs 1))))
(define start-time (runtime))

(write (go))
(newline)
(write (sqrt 2.0))
(newline)
(write (round 209.177))
(newline)

(run)

(##declare (generic))

(let ((i intrs) (elapsed (- (runtime) start-time)))
  (display elapsed ##stderr)
  (display " secs elapsed cpu time (user+system time)" ##stderr)
  (newline ##stderr)
  (if (or (< i 2)
          (< (* elapsed 11) (- i 1))
          (> (* elapsed 9) (+ i 1)))
    (begin
      (display "*** possible timer problem: " ##stderr)
      (write i ##stderr)
      (display " interrupts in " ##stderr)
      (write elapsed ##stderr)
      (display " secs" ##stderr)
      (newline ##stderr))))

(exit)
