;==============================================================================

; file: "_ptree1.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Parse tree manipulation module: (part 1)
; ------------------------------

; This module contains procedures to construct the parse tree of a Scheme
; expression and manipulate the parse tree.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Definition of the structures found in the parse tree.

; These structures define the nodes associated to expressions.

; information common to all nodes

;  parent   ; the node of which this node is a child
;  children ; list of parse-trees of the sub-expressions
;  fv       ; set of free/non-global vars contained in this expr
;  decl     ; declarations that apply to this node
;  source   ; source corresponding to this node

(define (node-parent x)          (vector-ref x 1))
(define (node-children x)        (vector-ref x 2))
(define (node-fv x)              (vector-ref x 3))
(define (node-decl x)            (vector-ref x 4))
(define (node-source x)          (vector-ref x 5))
(define (node-parent-set! x y)   (vector-set! x 1 y))
(define (node-fv-set! x y)       (vector-set! x 3 y))
(define (node-decl-set! x y)     (vector-set! x 4 y))
(define (node-source-set! x y)   (vector-set! x 5 y))

(define (node-children-set! x y)
  (vector-set! x 2 y)
  (for-each (lambda (child) (node-parent-set! child x)) y)
  (node-fv-invalidate! x))

(define (node-fv-invalidate! x)
  (let loop ((node x))
    (if node
      (begin
        (node-fv-set! node #t)
        (loop (node-parent node))))))

(define (make-cst ; node that represents constants
         parent children fv decl source ; common to all nodes

    val) ; value of the constant

  (vector cst-tag parent children fv decl source val))

(define (cst? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) cst-tag)))

(define (cst-val x)        (vector-ref x 6))
(define (cst-val-set! x y) (vector-set! x 6 y))

(define cst-tag (list 'cst-tag))

(define (make-ref ; node that represents variable references
         parent children fv decl source ; common to all nodes

    var) ; the variable which is referenced

  (vector ref-tag parent children fv decl source var))

(define (ref? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) ref-tag)))

(define (ref-var x)        (vector-ref x 6))
(define (ref-var-set! x y) (vector-set! x 6 y))

(define ref-tag (list 'ref-tag))

(define (make-set ; node that represents assignments (i.e. set! special forms)
         parent children fv decl source ; common to all nodes

    var) ; the variable which is assigned a value

  (vector set-tag parent children fv decl source var))

(define (set? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) set-tag)))

(define (set-var x)        (vector-ref x 6))
(define (set-var-set! x y) (vector-set! x 6 y))

(define set-tag (list 'set-tag))

(define (make-def ; node that represents toplevel definitions
         parent children fv decl source ; common to all nodes

    var) ; the global variable which is assigned a value

  (vector def-tag parent children fv decl source var))

(define (def? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) def-tag)))

(define (def-var x)        (vector-ref x 6))
(define (def-var-set! x y) (vector-set! x 6 y))

(define def-tag (list 'def-tag))

(define (make-tst ; node that represents conditionals (i.e. if special forms)
         parent children fv decl source ; common to all nodes

    )

  (vector tst-tag parent children fv decl source))

(define (tst? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) tst-tag)))

(define tst-tag (list 'tst-tag))

(define (make-conj ; node that represents conjunctions (i.e. and special forms)
         parent children fv decl source ; common to all nodes

    )

  (vector conj-tag parent children fv decl source))

(define (conj? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) conj-tag)))

(define conj-tag (list 'conj-tag))

(define (make-disj ; node that represents disjunctions (i.e. or special forms)
         parent children fv decl source ; common to all nodes

    )

  (vector disj-tag parent children fv decl source))

(define (disj? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) disj-tag)))

(define disj-tag (list 'disj-tag))

(define (make-prc ; node that represents procedures (i.e. lambda-expressions)
         parent children fv decl source ; common to all nodes

    name     ; name of this procedure (string) or #f if none
    c-name   ; name of associated C function (string) or #f if none
    parms    ; the list of parameter variables in order
    opts     ; the list of objects that the optional parameters default to
    keys     ; the list of keyword/default pairs
    rest?)   ; #t if the last parameter is a rest parameter

  (vector prc-tag parent children fv decl source
          name
          c-name
          parms
          opts
          keys
          rest?))

(define (prc? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) prc-tag)))

(define (prc-name x)          (vector-ref x 6))
(define (prc-c-name x)        (vector-ref x 7))
(define (prc-parms x)         (vector-ref x 8))
(define (prc-opts x)          (vector-ref x 9))
(define (prc-keys x)          (vector-ref x 10))
(define (prc-rest? x)         (vector-ref x 11))
(define (prc-name-set! x y)   (vector-set! x 6 y))
(define (prc-c-name-set! x y) (vector-set! x 7 y))
(define (prc-parms-set! x y)  (vector-set! x 8 y))
(define (prc-opts-set! x y)   (vector-set! x 9 y))
(define (prc-keys-set! x y)   (vector-set! x 10 y))
(define (prc-rest?-set! x y)  (vector-set! x 11 y))

(define prc-tag (list 'prc-tag))

(define (make-app ; node that represents procedure calls
         parent children fv decl source ; common to all nodes

    )

  (vector app-tag parent children fv decl source))

(define (app? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) app-tag)))

(define app-tag (list 'app-tag))

(define (make-fut ; node that represents future constructs
         parent children fv decl source ; common to all nodes

    )

  (vector fut-tag parent children fv decl source))

(define (fut? x)
  (and (vector? x)
       (> (vector-length x) 0)
       (eq? (vector-ref x 0) fut-tag)))

(define fut-tag (list 'fut-tag))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Procedures to create parse tree nodes and extract sub-nodes.

(define (new-cst source decl val)
  (make-cst #f '() #t decl source val))

(define (new-ref source decl var)
  (let ((node (make-ref #f '() #t decl source var)))
    (var-refs-set! var (ptset-adjoin (var-refs var) node))
    node))

(define (new-ref-extended-bindings source name env)
  (new-ref source
           (add-extended-bindings (env-declarations env))
           (env-lookup-global-var env name)))

(define (new-set source decl var val)
  (let ((node (make-set #f (list val) #t decl source var)))
    (var-sets-set! var (ptset-adjoin (var-sets var) node))
    (node-parent-set! val node)
    node))

(define (set-val x)
  (if (set? x)
    (car (node-children x))
    (compiler-internal-error "set-val, 'set' node expected" x)))

(define (new-def source decl var val)
  (let ((node (make-def #f (list val) #t decl source var)))
    (var-sets-set! var (ptset-adjoin (var-sets var) node))
    (node-parent-set! val node)
    node))

(define (def-val x)
  (if (def? x)
    (car (node-children x))
    (compiler-internal-error "def-val, 'def' node expected" x)))

(define (new-tst source decl pre con alt)
  (let ((node (make-tst #f (list pre con alt) #t decl source)))
    (node-parent-set! pre node)
    (node-parent-set! con node)
    (node-parent-set! alt node)
    node))

(define (tst-pre x)
  (if (tst? x)
    (car (node-children x))
    (compiler-internal-error "tst-pre, 'tst' node expected" x)))

(define (tst-con x)
  (if (tst? x)
    (cadr (node-children x))
    (compiler-internal-error "tst-con, 'tst' node expected" x)))

(define (tst-alt x)
  (if (tst? x)
    (caddr (node-children x))
    (compiler-internal-error "tst-alt, 'tst' node expected" x)))

(define (new-conj source decl pre alt)
  (let ((node (make-conj #f (list pre alt) #t decl source)))
    (node-parent-set! pre node)
    (node-parent-set! alt node)
    node))

(define (conj-pre x)
  (if (conj? x)
    (car (node-children x))
    (compiler-internal-error "conj-pre, 'conj' node expected" x)))

(define (conj-alt x)
  (if (conj? x)
    (cadr (node-children x))
    (compiler-internal-error "conj-alt, 'conj' node expected" x)))

(define (new-disj source decl pre alt)
  (let ((node (make-disj #f (list pre alt) #t decl source)))
    (node-parent-set! pre node)
    (node-parent-set! alt node)
    node))

(define (disj-pre x)
  (if (disj? x)
    (car (node-children x))
    (compiler-internal-error "disj-pre, 'disj' node expected" x)))

(define (disj-alt x)
  (if (disj? x)
    (cadr (node-children x))
    (compiler-internal-error "disj-alt, 'disj' node expected" x)))

(define (new-prc source decl name c-name parms opts keys rest? body)
  (let* ((children (list body))
         (node (make-prc #f children #t decl source
                         name c-name parms opts keys rest?)))
    (for-each (lambda (x) (var-bound-set! x node)) parms)
    (node-parent-set! body node)
    node))

(define (prc-body x)
  (if (prc? x)
    (car (node-children x))
    (compiler-internal-error "prc-body, 'proc' node expected" x)))

(define (prc-req-parms-only? x)
  (and (null? (prc-opts x))
       (not (prc-keys x))
       (not (prc-rest? x))))

(define (new-call source decl oper args)
  (let ((node (make-app #f (cons oper args) #t decl source)))
    (node-parent-set! oper node)
    (for-each (lambda (x) (node-parent-set! x node)) args)
    node))

(define (new-call* source decl oper args)
  (new-call source decl oper args))

(define (app-oper x)
  (if (app? x)
    (car (node-children x))
    (compiler-internal-error "app-oper, 'call' node expected" x)))

(define (app-args x)
  (if (app? x)
    (cdr (node-children x))
    (compiler-internal-error "app-args, 'call' node expected" x)))

(define (oper-pos? node)
  (let ((parent (node-parent node)))
    (if parent
      (and (app? parent)
           (eq? (app-oper parent) node))
      #f)))

(define (new-fut source decl val)
  (let ((node (make-fut #f (list val) #t decl source)))
    (node-parent-set! val node)
    node))

(define (fut-val x)
  (if (fut? x)
    (car (node-children x))
    (compiler-internal-error "fut-val, 'fut' node expected" x)))

(define (new-disj-call source decl pre oper alt)
  (new-call* source decl
    (let* ((temp (new-temp-variable source 'cond-temp))
           (parms (list temp)))
      (new-prc source decl #f #f parms '() #f #f
        (new-tst source decl
          (new-ref source decl temp)
          (new-call* source decl oper (list (new-ref source decl temp)))
          alt)))
    (list pre)))

(define (new-seq source decl before after)
  (let ((temp (new-temp-variable source 'begin-temp)))
    (new-call* source decl
      (new-prc source decl #f #f (list temp) '() #f #f
        after)
      (list before))))

(define (new-let ptree proc vars vals body)
  (if (pair? vars)
    (new-call (node-source ptree) (node-decl ptree)
      (new-prc (node-source proc) (node-decl proc)
        (prc-name proc)
        (prc-c-name proc)
        (reverse vars)
        '()
        #f
        #f
        body)
      (reverse vals))
    body))

(define (new-temp-variable source name)
  (make-var name #t (ptset-empty) (ptset-empty) source))

(define (new-variables sources)
  (map new-variable sources))

(define (new-variable source)
  (make-var (source-code source) #t (ptset-empty) (ptset-empty) source))

(define (set-prc-names! vars vals)
  (let loop ((vars vars) (vals vals))
    (if (not (null? vars))
      (let ((var (car vars))
            (val (car vals)))
        (if (prc? val)
          (prc-name-set! val (symbol->string (var-name var))))
        (loop (cdr vars) (cdr vals))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Procedures to get variable classes from nodes.

(define (free-variables node) ; set of free variables used in the expression
  (if (eq? (node-fv node) #t)
    (let ((x (varset-union-multi (map free-variables (node-children node)))))
      (node-fv-set! node
        (cond ((ref? node)
               (if (global? (ref-var node))
                 x
                 (varset-adjoin x (ref-var node))))
              ((set? node)
               (if (global? (set-var node))
                 x
                 (varset-adjoin x (set-var node))))
              ((prc? node)
               (varset-difference x (bound-variables node)))
              ((and (app? node) (prc? (app-oper node)))
               (varset-difference x (bound-variables (app-oper node))))
              (else
               x)))))
  (node-fv node))

(define (bound-variables node) ; set of variables bound by a procedure
  (list->varset (prc-parms node)))

(define (not-mutable? var)
  (ptset-empty? (var-sets var)))

(define (mutable? var)
  (not (not-mutable? var)))

(define (bound? var)
  (var-bound var))

(define (global? var)
  (not (bound? var)))

(define (global-single-def var) ; get definition of a global if it is only
  (and (global? var)            ; defined once and it will never change
       (let ((sets (ptset->list (var-sets var))))
         (and (pair? sets)
              (null? (cdr sets))
              (def? (car sets))
              (block-compilation? (node-decl (car sets)))
              (def-val (car sets))))))

(define (global-proc-obj node)
  (let ((var (ref-var node)))
    (and (global? var)
         (standard-proc-obj (var-name var) (node-decl node)))))

(define (global-singly-bound? node)
  (or (global-single-def (ref-var node))
      (global-proc-obj node)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Declarations.

; Dialect related declarations:
;
; (ieee-scheme)     use IEEE Scheme
; (r4rs-scheme)     use R4RS Scheme
; (multilisp)       use Multilisp
;
; Lambda-lifting declarations:
;
; (lambda-lift)     can lambda-lift user procedures
; (not lambda-lift) can't lambda-lift user procedures
;
; Inlining declarations:
;
; (inline)                            compiler may inline user procedures
; (not inline)                        no user procedure will be inlined
;
; (inlining-limit n)                  inlined user procedures must not be
;                                     bigger than 'n'
;
; Compilation strategy declarations:
;
; (block)     global vars defined are only mutated by code in the current file
; (separate)  global vars defined can be mutated by other code
;
; Global variable binding declarations:
;
; (standard-bindings)                 compiler can assume standard bindings
; (standard-bindings <var1> ...)      assume st. bind. for vars specified
; (not standard-bindings)             can't assume st. bind. for any var
; (not standard-bindings <var1> ...)  can't assume st. bind. for vars spec.
;
; (extended-bindings)                 compiler can assume extended bindings
; (extended-bindings <var1> ...)      assume ext. bind. for vars specified
; (not extended-bindings)             can't assume ext. bind. for any var
; (not extended-bindings <var1> ...)  can't assume ext. bind. for vars spec.
;
; Code safety declarations:
;
; (safe)                              runtime errors won't crash system
; (not safe)                          assume program doesn't contain errors
;
; Interrupt checking declarations:
;
; (interrupts-enabled)                allow interrupts
; (not interrupts-enabled)            disallow interrupts

(define-flag-decl ieee-scheme-sym 'dialect)
(define-flag-decl r4rs-scheme-sym 'dialect)
(define-flag-decl multilisp-sym   'dialect)

(define-boolean-decl lambda-lift-sym)

(define-boolean-decl inline-sym)
(define-parameterized-decl inlining-limit-sym)

(define-flag-decl block-sym    'compilation-strategy)
(define-flag-decl separate-sym 'compilation-strategy)

(define-namable-boolean-decl standard-bindings-sym)
(define-namable-boolean-decl extended-bindings-sym)

(define-boolean-decl safe-sym)

(define-boolean-decl interrupts-enabled-sym)

(define (scheme-dialect decl) ; returns dialect in effect
  (declaration-value 'dialect #f ieee-scheme-sym decl))

(define (lambda-lift? decl) ; true iff should lambda-lift
  (declaration-value lambda-lift-sym #f #t decl))

(define (inline? decl) ; true iff should inline
  (declaration-value inline-sym #f #t decl))

(define (inlining-limit decl) ; returns the inlining limit
  (max 0 (min 10000 (declaration-value inlining-limit-sym #f 300 decl))))

(define (block-compilation? decl) ; true iff block compilation strategy
  (eq? (declaration-value 'compilation-strategy #f separate-sym decl)
       block-sym))

(define (standard-binding? name decl) ; true iff name's binding is standard
  (declaration-value standard-bindings-sym name #f decl))

(define (extended-binding? name decl) ; true iff name's binding is extended
  (declaration-value extended-bindings-sym name #f decl))

(define (add-extended-bindings decl)
  (add-decl (list extended-bindings-sym #t) decl))

(define (intrs-enabled? decl) ; true iff system should generate interrupt checks
  (declaration-value interrupts-enabled-sym #f #t decl))

(define (add-not-interrupts-enabled decl)
  (add-decl (list interrupts-enabled-sym #f) decl))

(define (safe? decl) ; true iff system should prevent fatal runtime errors
  (declaration-value safe-sym #f #t decl))

(define (add-not-safe decl)
  (add-decl (list safe-sym #f) decl))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Dialect info:

(define (dialect-specific-reserved-identifiers dialect)
  (cond ((eq? dialect ieee-scheme-sym)
         ieee-scheme-specific-reserved-identifiers)
        ((eq? dialect r4rs-scheme-sym)
         r4rs-scheme-specific-reserved-identifiers)
        (else
         multilisp-specific-reserved-identifiers)))

; IEEE Scheme

(define ieee-scheme-specific-reserved-identifiers
  '())

; R4RS Scheme

(define r4rs-scheme-specific-reserved-identifiers
  (list delay-sym))

; Multilisp

(define multilisp-specific-reserved-identifiers
  (list delay-sym future-sym))

; common stuff

(define common-reserved-identifiers
  (list quote-sym quasiquote-sym unquote-sym unquote-splicing-sym
        lambda-sym if-sym set!-sym cond-sym =>-sym else-sym and-sym or-sym
        case-sym let-sym let*-sym letrec-sym begin-sym do-sym define-sym
        **define-macro-sym **declare-sym **include-sym
        c-define-type-sym c-declare-sym c-initialize-sym c-lambda-sym
        c-define-sym))

(define (standard-proc-obj name decl)
  (let ((proc (target.prim-info name)))
    (and proc
         (standard-procedure?
           proc
           (standard-binding? name decl)
           (extended-binding? name decl)
           (scheme-dialect decl))
         proc)))

(define (standard-procedure? proc std? ext? dialect)
  (let ((standard (proc-obj-standard proc)))
    (if (eq? standard 'extended)
      ext?
      (and std?
           (cond ((eq? standard 'ieee)
                  #t)
                 ((eq? standard 'r4rs)
                  (not (eq? dialect ieee-scheme-sym)))
                 ((eq? standard 'multilisp)
                  (eq? dialect multilisp-sym))
                 (else
                  #f))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (parse-program program env module-name proc) returns a list of parse
; trees/environment pairs describing the program and the final global
; environment.

(define (parse-program program env module-name proc)

  (define (parse-prog program env lst proc)
    (if (null? program)
      (proc (reverse lst) env)
      (let ((source (car program)))

        (cond ((macro-expr? source env)
               (parse-prog
                 (cons (macro-expand source env) (cdr program))
                 env
                 lst
                 proc))

              ((begin-defs-expr? source)
               (parse-prog
                 (append (begin-defs-body source) (cdr program))
                 env
                 lst
                 proc))

              ((include-expr? source)

               (if *ptree-port*
                 (display "  " *ptree-port*))

               (let ((x (file->sources* (include-filename source)
                                        *ptree-port*
                                        (source-locat source))))

                 (if *ptree-port*
                   (newline *ptree-port*))
                      
                 (parse-prog
                   (append x (cdr program))
                   env
                   lst
                   proc)))

              ((define-macro-expr? source env)

               (if *ptree-port*
                 (begin
                   (display "  \"macro\"" *ptree-port*)
                   (newline *ptree-port*)))

               (parse-prog
                 (cdr program)
                 (add-macro source env)
                 lst
                 proc))

              ((declare-expr? source)

               (if *ptree-port*
                 (begin
                   (display "  \"decl\"" *ptree-port*)
                   (newline *ptree-port*)))

               (parse-prog
                 (cdr program)
                 (add-declarations source env)
                 lst
                 proc))

              ((define-expr? source env)
               (let* ((var-source (definition-variable source))
                      (var (source-code var-source))
                      (v (env-lookup-var env var var-source)))

                 (if *ptree-port*
                   (begin
                     (display "  " *ptree-port*)
                     (write (var-name v) *ptree-port*)
                     (newline *ptree-port*)))

                 (let ((node (pt (definition-value source) env 'true)))
                   (set-prc-names! (list v) (list node))
                   (parse-prog
                     (cdr program)
                     env
                     (cons (cons (new-def source
                                          (env-declarations env)
                                          v
                                          node)
                                 env)
                           lst)
                     proc))))

              ((c-define-type-expr? source)
               (let ((name (source-code (c-type-definition-name source)))
                     (type-source (c-type-definition-type source)))

                 (if *ptree-port*
                   (begin
                     (display "  \"c-define-type\"" *ptree-port*)
                     (newline *ptree-port*)))

                 (add-c-type name type-source)

                 (parse-prog
                   (cdr program)
                   env
                   lst
                   proc)))

              ((c-declare-expr? source)
               (let ((body (source-code (c-declaration-body source))))

                 (if *ptree-port*
                   (begin
                     (display "  \"c-declare\"" *ptree-port*)
                     (newline *ptree-port*)))

                 (add-c-decl body)

                 (parse-prog
                   (cdr program)
                   env
                   lst
                   proc)))

              ((c-initialize-expr? source)
               (let ((body (source-code (c-initialization-body source))))

                 (if *ptree-port*
                   (begin
                     (display "  \"c-initialize\"" *ptree-port*)
                     (newline *ptree-port*)))

                 (add-c-init body)

                 (parse-prog
                   (cdr program)
                   env
                   lst
                   proc)))

              ((c-define-expr? source env)
               (let* ((var-source (c-definition-variable source))
                      (var (source-code var-source))
                      (v (env-lookup-var env var var-source))
                      (param-types (c-definition-param-types source))
                      (result-type (c-definition-result-type source))
                      (proc-name-source (c-definition-proc-name source))
                      (proc-name (source-code proc-name-source))
                      (scope-source (c-definition-scope source))
                      (scope (source-code scope-source)))

                 (if *ptree-port*
                   (begin
                     (display "  " *ptree-port*)
                     (write (var-name v) *ptree-port*)
                     (newline *ptree-port*)))

                 (build-c-define param-types result-type proc-name scope)

                 (let ((node (pt (c-definition-value source) env 'true)))
                   (set-prc-names! (list v) (list node))
                   (prc-c-name-set! node proc-name)
                   (parse-prog
                     (cdr program)
                     env
                     (cons (cons (new-def source
                                          (env-declarations env)
                                          v
                                          node)
                                 env)
                           lst)
                     proc))))

              (else

               (if *ptree-port*
                 (begin
                   (display "  \"expr\"" *ptree-port*)
                   (newline *ptree-port*)))

               (parse-prog
                 (cdr program)
                 env
                 (cons (cons (pt source env 'true) env) lst)
                 proc))))))

  (if *ptree-port*
    (begin
      (display "Parsing:" *ptree-port*)
      (newline *ptree-port*)))

  (c-interface-begin module-name)

  (parse-prog program env '()
    (lambda (lst env)

      (if *ptree-port*
        (newline *ptree-port*))

      (check-multiple-global-defs env)

      (proc lst env (c-interface-end)))))

(define (check-multiple-global-defs env)
  (let ((global-vars (env-global-variables env)))
    (for-each
      (lambda (var)
        (let ((defs (keep def? (ptset->list (var-sets var)))))
          (if (> (length defs) 1)
            (begin
              (display "*** WARNING -- More than one 'define' of global variable \"")
              (display (var-name var))
              (display "\".")
              (newline)))))
      global-vars)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; (pt source env use) returns the parse tree for the Scheme source expression
; 'source' in the environment 'env'.  If 'source' is not syntactically
; correct, an error is signaled.  The value of 'use' determines what the
; expression's value will be used for; it must be one of the following:
;
;  true  : the true value of the expression is needed
;  pred  : the value is used as a predicate
;  none  : the value is not needed (but its side effect might)

(define (pt-syntax-error source msg . args)
  (apply compiler-user-error
         (cons (source-locat source)
               (cons (string-append "Syntax error -- " msg)
                     args))))

(define (pt source env use)
  (cond ((macro-expr? source env)        (pt (macro-expand source env) env use))
        ((self-eval-expr? source)        (pt-self-eval source env use))
        ((quote-expr? source)            (pt-quote source env use))
        ((quasiquote-expr? source)       (pt-quasiquote source env use))
        ((unquote-expr? source)
         (pt-syntax-error source "Ill-placed 'unquote'"))
        ((unquote-splicing-expr? source)
         (pt-syntax-error source "Ill-placed 'unquote-splicing'"))
        ((var-expr? source env)          (pt-var source env use))
        ((set!-expr? source env)         (pt-set! source env use))
        ((lambda-expr? source env)       (pt-lambda source env use))
        ((if-expr? source)               (pt-if source env use))
        ((cond-expr? source)             (pt-cond source env use))
        ((and-expr? source)              (pt-and source env use))
        ((or-expr? source)               (pt-or source env use))
        ((case-expr? source)             (pt-case source env use))
        ((let-expr? source env)          (pt-let source env use))
        ((let*-expr? source env)         (pt-let* source env use))
        ((letrec-expr? source env)       (pt-letrec source env use))
        ((begin-expr? source)            (pt-begin source env use))
        ((do-expr? source env)           (pt-do source env use))
        ((define-expr? source env)
         (pt-syntax-error source "Ill-placed 'define'"))
        ((delay-expr? source env)        (pt-delay source env use))
        ((future-expr? source env)       (pt-future source env use))
        ((define-macro-expr? source env)
         (pt-syntax-error source "Ill-placed '##define-macro'"))
        ((begin-defs-expr? source)
         (pt-syntax-error source "Ill-placed 'begin' style definitions"))
        ((declare-expr? source)
         (pt-syntax-error source "Ill-placed '##declare'"))
        ((c-define-type-expr? source)
         (pt-syntax-error source "Ill-placed 'c-define-type'"))
        ((c-declare-expr? source)
         (pt-syntax-error source "Ill-placed 'c-declare'"))
        ((c-initialize-expr? source)
         (pt-syntax-error source "Ill-placed 'c-initialize'"))
        ((c-lambda-expr? source)         (pt-c-lambda source env use))
        ((c-define-expr? source env)
         (pt-syntax-error source "Ill-placed 'c-define'"))
        ((combination-expr? source)      (pt-combination source env use))
        (else
         (pt-syntax-error source "Illegal expression"))))

(define (macro-expand source env)
  (let ((code (source-code source)))
    (expression->source
      (apply (cdr (env-lookup-macro env (source-code (car code))))
             (cdr (source->expression source)))
      source)))

(define (pt-self-eval source env use)
  (let ((val (source->expression source)))
    (if (eq? use 'none)
      (new-cst source (env-declarations env) void-object)
      (new-cst source (env-declarations env) val))))

(define (pt-quote source env use)
  (let ((code (source-code source)))
    (if (eq? use 'none)
      (new-cst source (env-declarations env) void-object)
      (new-cst source (env-declarations env) (source->expression (cadr code))))))

(define (pt-quasiquote source env use)
  (let ((code (source-code source)))
    (pt-quasiquotation (cadr code) 1 env)))

(define (pt-quasiquotation form level env)
  (cond ((= level 0)
         (pt form env 'true))
        ((quasiquote-expr? form)
         (pt-quasiquotation-list form (source-code form) (+ level 1) env))
        ((unquote-expr? form)
         (if (= level 1)
           (pt (cadr (source-code form)) env 'true)
           (pt-quasiquotation-list form (source-code form) (- level 1) env)))
        ((unquote-splicing-expr? form)
         (if (= level 1)
           (pt-syntax-error form "Ill-placed 'unquote-splicing'")
           (pt-quasiquotation-list form (source-code form) (- level 1) env)))
        ((pair? (source-code form))
         (pt-quasiquotation-list form (source-code form) level env))
        ((vector? (source-code form))
         (let ((lst (vect->list (source-code form))))
           (vector-form
             form
             (pt-quasiquotation-list form lst level env)
             env)))
        (else
         (new-cst form (env-declarations env) (source->expression form)))))

(define (pt-quasiquotation-list form l level env)
  (cond ((pair? l)
         (if (and (unquote-splicing-expr? (car l)) (= level 1))
           (let ((x (pt (cadr (source-code (car l))) env 'true)))
             (if (null? (cdr l))
               x
               (append-form (car l)
                            x
                            (pt-quasiquotation-list form (cdr l) 1 env)
                            env)))
           (cons-form form
                      (pt-quasiquotation (car l) level env)
                      (pt-quasiquotation-list form (cdr l) level env)
                      env)))
        ((null? l)
         (new-cst form (env-declarations env) '()))
        (else
         (pt-quasiquotation l level env))))

(define (append-form source ptree1 ptree2 env)
  (cond ((and (cst? ptree1) (cst? ptree2))
         (new-cst source (env-declarations env)
           (append (cst-val ptree1) (cst-val ptree2))))
        ((and (cst? ptree2) (null? (cst-val ptree2)))
         ptree1)
        (else
         (new-call* source (add-not-safe (env-declarations env))
           (new-ref-extended-bindings source **quasi-append-sym env)
           (list ptree1 ptree2)))))

(define (cons-form source ptree1 ptree2 env)
  (cond ((and (cst? ptree1) (cst? ptree2))
         (new-cst source (env-declarations env)
           (cons (cst-val ptree1) (cst-val ptree2))))
        ((and (cst? ptree2) (null? (cst-val ptree2)))
         (new-call* source (add-not-safe (env-declarations env))
           (new-ref-extended-bindings source **quasi-list-sym env)
           (list ptree1)))
        (else
         (new-call* source (add-not-safe (env-declarations env))
           (new-ref-extended-bindings source **quasi-cons-sym env)
           (list ptree1 ptree2)))))

(define (vector-form source ptree env)
  (if (cst? ptree)
    (new-cst source (env-declarations env)
      (list->vect (cst-val ptree)))
    (new-call* source (add-not-safe (env-declarations env))
      (new-ref-extended-bindings source **quasi-list->vector-sym env)
      (list ptree))))

(define (pt-var source env use)
  (if (eq? use 'none)
    (new-cst source (env-declarations env) void-object)
    (new-ref source (env-declarations env)
      (env-lookup-var env (source-code source) source))))

(define (pt-set! source env use)
  (let ((code (source-code source)))
    (new-set source (env-declarations env)
      (env-lookup-var env (source-code (cadr code)) (cadr code))
      (pt (caddr code) env 'true))))

(define (pt-lambda source env use)

  (define (check-none-result node)
    (if (eq? use 'none)
      (new-cst source (env-declarations env) void-object)
      node))

  (define (bind-default-bindings default-bindings env)
    (if (null? default-bindings)
      (pt-body source (cddr (source-code source)) env 'true)
      (let* ((binding (car default-bindings))
             (var1 (vector-ref binding 0))
             (var2 (vector-ref binding 1))
             (val (vector-ref binding 2))
             (parm-source (vector-ref binding 3))
             (vars (list var2))
             (decl (env-declarations env)))
        (new-call* parm-source decl
          (new-prc parm-source decl #f #f vars '() #f #f
            (bind-default-bindings
              (cdr default-bindings)
              (env-frame env vars)))
          (list (new-tst parm-source decl
                  (new-call* parm-source decl
                    (new-ref-extended-bindings
                      parm-source
                      **eq?-sym
                      env)
                    (list (new-ref parm-source decl var1)
                          (new-cst parm-source decl absent-object)))
                  val
                  (new-ref parm-source decl var1)))))))

  (define (split-default-bindings parms env cont)
    (let loop ((lst parms)
               (rev-vars '())
               (rev-defaults '())
               (rev-bindings '())
               (env env))
      (if (null? lst)

        (cont (reverse rev-vars)
              (reverse rev-defaults)
              (reverse rev-bindings)
              env)

        (let* ((parameter
                (car lst))
               (parm-source
                (parameter-source parameter))
               (val-source
                (parameter-default-source parameter))
               (var1
                (new-variable parm-source))
               (val
                (if val-source
                  (pt val-source env 'true)
                  (new-cst parm-source (env-declarations env)
                    false-object))))
          (if (cst? val)
            (loop (cdr lst)
                  (cons var1 rev-vars)
                  (cons (cst-val val) rev-defaults)
                  rev-bindings
                  (env-frame env (list var1)))
            (let ((var2 (new-variable parm-source)))
              (loop (cdr lst)
                    (cons var1 rev-vars)
                    (cons absent-object rev-defaults)
                    (cons (vector var1 var2 val parm-source)
                          rev-bindings)
                    (env-frame env (list var2)))))))))

  (let* ((code
          (source-code source))
         (all-parms
          (extract-parameters (source->parms (cadr code)) env))
         (required-parameters
          (vector-ref all-parms 0))
         (optional-parameters
          (vector-ref all-parms 1))
         (rest-parameter
          (vector-ref all-parms 2))
         (key-parameters
          (vector-ref all-parms 3))
         (required-vars
          (new-variables (map parameter-source required-parameters)))
         (rest-vars
          (if rest-parameter
            (list (new-variable (parameter-source rest-parameter)))
            '())))

    (check-none-result
     (split-default-bindings
      (or optional-parameters '())
      (env-frame env required-vars)
      (lambda (opt-vars opt-defaults opt-bindings opt-env)
        (split-default-bindings
         (or key-parameters '())
         (env-frame opt-env rest-vars)
         (lambda (key-vars key-defaults key-bindings key-env)
           (let ((keys
                  (and key-parameters
                       (map (lambda (x)
                              (cons (string->keyword-object
                                     (symbol->string (var-name (car x))))
                                    (cdr x)))
                            (pair-up key-vars key-defaults))))
                 (outer-vars
                  (append required-vars opt-vars key-vars rest-vars))
                 (decl
                  (env-declarations env)))
             (new-prc source decl #f #f outer-vars opt-defaults keys
               (not (not rest-parameter)) ; turn to boolean
               (bind-default-bindings
                (append opt-bindings key-bindings)
                (env-frame env outer-vars)))))))))))

;***************
;(let ((z (extract-parameters (source->parms (cadr code)) env)))
;(pp (list (map (lambda (x) (vector-ref x 0)) (vector-ref z 0))
;          (map (lambda (x) (vector (vector-ref x 0) (if (vector-ref x 2) (source->expression (vector-ref x 2)) #f))) (vector-ref z 1))
;          (if (vector-ref z 2) (vector-ref (vector-ref z 2) 0) #f)
;          (map (lambda (x) (vector (vector-ref x 0) (if (vector-ref x 2) (source->expression (vector-ref x 2)) #f))) (vector-ref z 3))))

(define (parameter-name parm)
  (vector-ref parm 0))

(define (parameter-source parm)
  (vector-ref parm 1))

(define (parameter-default-source parm)
  (vector-ref parm 2))

(define (extract-parameters param-list env)

  (define (parm-expected-err source)
    (pt-syntax-error source "Parameter must be an identifier"))

  (define (parm-or-default-binding-expected-err source)
    (pt-syntax-error source "Parameter must be an identifier or default binding"))

  (define (duplicate-parm-err source)
    (pt-syntax-error source "Duplicate parameter in parameter list"))

  (define (dotted-rest-err source)
    (pt-syntax-error source "Ill-placed dotted rest parameter"))

  (define (rest-parm-expected-err source)
    (pt-syntax-error source "#!rest must be followed by a parameter"))

  (define (default-binding-err source)
    (pt-syntax-error source "Ill-formed default binding"))

  (define (optional-illegal-err source)
    (pt-syntax-error source "Ill-placed #!optional"))

  (define (rest-illegal-err source)
    (pt-syntax-error source "Ill-placed #!rest"))

  (define (key-illegal-err source)
    (pt-syntax-error source "Ill-placed #!key"))

  (define (key-expected-err source)
    (pt-syntax-error source "#!key expected after rest parameter"))

  (define (default-binding-illegal-err source)
    (pt-syntax-error source "Ill-placed default binding"))

  (let loop ((lst param-list)
             (rev-required-parms '())
             (rev-optional-parms #f)
             (rest-parm #f)
             (rev-key-parms #f)
             (state 1)) ; 1 = required parms or #!optional or #!rest or #!key
                        ; 2 = optional parms or #!rest or #!key
                        ; 3 = #!key
                        ; 4 = key parms

    (define (done rest-parm)
      (vector (reverse rev-required-parms)
              (and rev-optional-parms (reverse rev-optional-parms))
              rest-parm
              (if (or (not rev-key-parms)
                      (and (null? rev-key-parms) (not rest-parm)))
                #f
                (reverse rev-key-parms))))

    (define (parm-exists? parm lst)
      (and lst
           (not (null? lst))
           (or (eq? parm (vector-ref (car lst) 0))
               (parm-exists? parm (cdr lst)))))

    (define (check-if-duplicate parm parm-source)
      (if (or (parm-exists? parm rev-required-parms)
              (parm-exists? parm rev-optional-parms)
              (and rest-parm (eq? parm (vector-ref rest-parm 0)))
              (parm-exists? parm rev-key-parms))
        (duplicate-parm-err parm-source)))

    (cond ((null? lst)
           (done rest-parm))
          ((pair? lst)
           (let* ((parm-source (car lst))
                  (parm (source-code parm-source)))
             (cond ((optional-object? parm)
                    (if (> state 1)
                      (optional-illegal-err parm-source))
                    (loop (cdr lst)
                          rev-required-parms
                          '()
                          #f
                          #f
                          2))
                   ((rest-object? parm)
                    (if (> state 2)
                      (rest-illegal-err parm-source))
                    (if (pair? (cdr lst))
                      (let* ((parm-source (cadr lst))
                             (parm (source-code parm-source)))
                        (if (bindable-var? parm-source env)
                          (begin
                            (check-if-duplicate parm parm-source)
                            (loop (cddr lst)
                                  rev-required-parms
                                  rev-optional-parms
                                  (vector parm parm-source)
                                  #f
                                  3))
                          (parm-expected-err parm-source)))
                      (rest-parm-expected-err parm-source)))
                   ((key-object? parm)
                    (if (> state 3)
                      (key-illegal-err parm-source))
                    (loop (cdr lst)
                          rev-required-parms
                          rev-optional-parms
                          rest-parm
                          '()
                          4))
                   ((= state 3)
                    (key-expected-err parm-source))
                   ((bindable-var? parm-source env)
                    (check-if-duplicate parm parm-source)
                    (case state
                      ((1)
                       (loop (cdr lst)
                             (cons (vector parm parm-source)
                                   rev-required-parms)
                             #f
                             #f
                             #f
                             1))
                      ((2)
                       (loop (cdr lst)
                             rev-required-parms
                             (cons (vector parm parm-source #f)
                                   rev-optional-parms)
                             #f
                             #f
                             2))
                      (else
                       (loop (cdr lst)
                             rev-required-parms
                             rev-optional-parms
                             rest-parm
                             (cons (vector parm parm-source #f)
                                   rev-key-parms)
                             4))))
                   ((pair? parm)
                    (if (not (or (= state 2) (= state 4)))
                      (default-binding-illegal-err parm-source))
                    (let ((length (proper-length parm)))
                      (if (not (eqv? length 2))
                        (default-binding-err parm-source)))
                    (let* ((parm-source (car parm))
                           (val-source (cadr parm))
                           (parm (source-code parm-source)))
                      (if (bindable-var? parm-source env)
                        (begin
                          (check-if-duplicate parm parm-source)
                          (case state
                            ((2)
                             (loop (cdr lst)
                                   rev-required-parms
                                   (cons (vector parm parm-source val-source)
                                         rev-optional-parms)
                                   #f
                                   #f
                                   2))
                            (else
                             (loop (cdr lst)
                                   rev-required-parms
                                   rev-optional-parms
                                   rest-parm
                                   (cons (vector parm parm-source val-source)
                                         rev-key-parms)
                                   4))))
                        (parm-expected-err parm-source))))
                   (else
                    (if (> state 1)
                      (parm-or-default-binding-expected-err parm-source)
                      (parm-expected-err parm-source))))))
          (else
           (let* ((parm-source lst)
                  (parm (source-code parm-source)))
             (if (bindable-var? parm-source env)
               (begin
                 (if (or rev-optional-parms
                         rest-parm
                         rev-key-parms)
                   (dotted-rest-err parm-source))
                 (check-if-duplicate parm parm-source)
                 (done (vector parm parm-source)))
               (parm-expected-err parm-source)))))))

(define (source->parms source)
  (let ((x (source-code source)))
    (if (or (pair? x) (null? x)) x source)))

(define (pt-body source body env use)

  (define (letrec-defines vars vals envs body env)
    (cond ((null? body)
           (pt-syntax-error
             source
             "Body must contain at least one evaluable expression"))
          ((macro-expr? (car body) env)
           (letrec-defines vars
                           vals
                           envs
                           (cons (macro-expand (car body) env)
                                 (cdr body))
                           env))
          ((begin-defs-expr? (car body))
           (letrec-defines vars
                           vals
                           envs
                           (append (begin-defs-body (car body))
                                   (cdr body))
                           env))
          ((include-expr? (car body))
           (if *ptree-port*
             (display "  " *ptree-port*))
           (let ((x (file->sources* (include-filename (car body))
                                    *ptree-port*
                                    (source-locat (car body)))))
             (if *ptree-port*
               (newline *ptree-port*))
             (letrec-defines vars
                             vals
                             envs
                             (append x (cdr body))
                             env)))
          ((define-expr? (car body) env)
           (let* ((var-source (definition-variable (car body)))
                  (var (source-code var-source))
                  (v (env-define-var env var var-source)))
             (letrec-defines (cons v vars)
                             (cons (definition-value (car body)) vals)
                             (cons env envs)
                             (cdr body)
                             env)))
          ((declare-expr? (car body))
           (letrec-defines vars
                           vals
                           envs
                           (cdr body)
                           (add-declarations (car body) env)))
          ((define-macro-expr? (car body) env)
           (letrec-defines vars
                           vals
                           envs
                           (cdr body)
                           (add-macro (car body) env)))
          ((null? vars)
           (pt-sequence source body env use))
          (else
           (let ((vars* (reverse vars)))
             (let loop ((vals* '()) (l1 vals) (l2 envs))
               (if (not (null? l1))
                 (loop (cons (pt (car l1) (car l2) 'true) vals*)
                       (cdr l1)
                       (cdr l2))
                 (pt-recursive-let source vars* vals* body env use)))))))

  (letrec-defines '() '() '() body (env-frame env '())))

(define (pt-sequence source seq env use)
  (if (length? seq 1)
    (pt (car seq) env use)
    (new-seq source (env-declarations env)
      (pt (car seq) env 'none)
      (pt-sequence source (cdr seq) env use))))

(define (pt-if source env use)
  (let ((code (source-code source)))
    (new-tst source (env-declarations env)
      (pt (cadr code) env 'pred)
      (pt (caddr code) env use)
      (if (length? code 3)
        (new-cst source (env-declarations env) void-object)
        (pt (cadddr code) env use)))))

(define (pt-cond source env use)

  (define (pt-clauses clauses)
    (if (length? clauses 0)
      (new-cst source (env-declarations env) void-object)
      (let* ((clause-source (car clauses))
             (clause (source-code clause-source)))
        (cond ((eq? (source-code (car clause)) else-sym)
               (pt-sequence clause-source (cdr clause) env use))
              ((length? clause 1)
               (new-disj clause-source (env-declarations env)
                 (pt (car clause) env (if (eq? use 'true) 'true 'pred))
                 (pt-clauses (cdr clauses))))
              ((eq? (source-code (cadr clause)) =>-sym)
               (new-disj-call clause-source (env-declarations env)
                 (pt (car clause) env 'true)
                 (pt (caddr clause) env 'true)
                 (pt-clauses (cdr clauses))))
              (else
               (new-tst clause-source (env-declarations env)
                 (pt (car clause) env 'pred)
                 (pt-sequence clause-source (cdr clause) env use)
                 (pt-clauses (cdr clauses))))))))

  (pt-clauses (cdr (source-code source))))

(define (pt-and source env use)

  (define (pt-exprs exprs)
    (cond ((length? exprs 0)
           (new-cst source (env-declarations env) #t))
          ((length? exprs 1)
           (pt (car exprs) env use))
          (else
           (new-conj (car exprs) (env-declarations env)
             (pt (car exprs) env (if (eq? use 'true) 'true 'pred))
             (pt-exprs (cdr exprs))))))

  (pt-exprs (cdr (source-code source))))

(define (pt-or source env use)

  (define (pt-exprs exprs)
    (cond ((length? exprs 0)
           (new-cst source (env-declarations env) false-object))
          ((length? exprs 1)
           (pt (car exprs) env use))
          (else
           (new-disj (car exprs) (env-declarations env)
             (pt (car exprs) env (if (eq? use 'true) 'true 'pred))
             (pt-exprs (cdr exprs))))))

  (pt-exprs (cdr (source-code source))))

(define (pt-case source env use)
  (let ((code (source-code source))
        (temp (new-temp-variable source 'case-temp)))

    (define (pt-clauses clauses)
      (if (length? clauses 0)
        (new-cst source (env-declarations env) void-object)
        (let* ((clause-source (car clauses))
               (clause (source-code clause-source)))

          (define (pt-inlined-memv constants)
            (let ((test
                    (new-call*
                      clause-source
                      (add-not-safe (env-declarations env))
                      (new-ref-extended-bindings clause-source **eqv?-sym env)
                      (list (new-ref clause-source (env-declarations env)
                              temp)
                            (new-cst (car clause) (env-declarations env)
                              (car constants))))))
              (if (null? (cdr constants))
                test
                (new-disj clause-source (env-declarations env)
                  test
                  (pt-inlined-memv (cdr constants))))))


          (if (eq? (source-code (car clause)) else-sym)
            (pt-sequence clause-source (cdr clause) env use)
            (new-tst clause-source (env-declarations env)
              (pt-inlined-memv (source->expression (car clause)))
              (pt-sequence clause-source (cdr clause) env use)
              (pt-clauses (cdr clauses)))))))

    (new-call* source (env-declarations env)
      (new-prc source (env-declarations env) #f #f (list temp) '() #f #f
        (pt-clauses (cddr code)))
      (list (pt (cadr code) env 'true)))))

(define (pt-let source env use)
  (let ((code (source-code source)))
    (if (bindable-var? (cadr code) env)
      (let* ((self (list (new-variable (cadr code))))
             (bindings (map source-code (source-code (caddr code))))
             (vars (new-variables (map car bindings)))
             (vals (map (lambda (x) (pt (cadr x) env 'true)) bindings))
             (env  (env-frame (env-frame env vars) self))
             (self-proc (list (new-prc source (env-declarations env)
                                #f
                                #f
                                vars
                                '()
                                #f
                                #f
                                (pt-body source (cdddr code) env use)))))
        (set-prc-names! self self-proc)
        (set-prc-names! vars vals)
        (new-call* source (env-declarations env)
          (new-prc source (env-declarations env) #f #f self '() #f #f
            (new-call* source (env-declarations env)
              (new-ref source (env-declarations env) (car self))
              vals))
          self-proc))
      (if (null? (source-code (cadr code)))
        (pt-body source (cddr code) env use)
        (let* ((bindings (map source-code (source-code (cadr code))))
               (vars (new-variables (map car bindings)))
               (vals (map (lambda (x) (pt (cadr x) env 'true)) bindings))
               (env  (env-frame env vars)))
          (set-prc-names! vars vals)
          (new-call* source (env-declarations env)
            (new-prc source (env-declarations env)
              #f
              #f
              vars
              '()
              #f
              #f
              (pt-body source (cddr code) env use))
            vals))))))

(define (pt-let* source env use)
  (let ((code (source-code source)))

    (define (pt-bindings bindings env use)
      (if (null? bindings)
        (pt-body source (cddr code) env use)
        (let* ((binding-source (car bindings))
               (binding (source-code binding-source))
               (vars (list (new-variable (car binding))))
               (vals (list (pt (cadr binding) env 'true)))
               (decl (env-declarations env))
               (env  (env-frame env vars)))
          (set-prc-names! vars vals)
          (new-call* binding-source decl
            (new-prc binding-source decl #f #f vars '() #f #f
              (pt-bindings (cdr bindings) env use))
            vals))))

    (pt-bindings (source-code (cadr code)) env use)))

(define (pt-letrec source env use)
  (let* ((code (source-code source))
         (bindings (map source-code (source-code (cadr code))))
         (vars* (new-variables (map car bindings)))
         (env*  (env-frame env vars*)))
    (pt-recursive-let
      source
      vars*
      (map (lambda (x) (pt (cadr x) env* 'true)) bindings)
      (cddr code)
      env*
      use)))

(define (pt-recursive-let source vars vals body env use)

  (define (dependency-graph vars vals)
    (let ((var-set (list->varset vars)))

      (define (dgraph vars vals)
        (if (null? vars)
          '()
          (let ((var (car vars)) (val (car vals)))
            (cons (make-gnode var (varset-intersection
                                    var-set
                                    (free-variables val)))
                  (dgraph (cdr vars) (cdr vals))))))

      (dgraph vars vals)))

  (define (val-of var)
    (list-ref vals (- (length vars) (length (memq var vars)))))

  (define (bind-in-order order)
    (if (null? order)
      (pt-body source body env use)

      ; get vars to be bound and vars to be assigned

      (let* ((vars-set (car order))
             (vars (varset->list vars-set)))
        (let loop1 ((l (reverse vars)) (vars-b '()) (vals-b '()) (vars-a '()))
          (if (not (null? l))

            (let* ((var (car l))
                   (val (val-of var)))
              (if (or (prc? val)
                      (not (varset-intersects? (free-variables val) vars-set)))
                (loop1 (cdr l)
                       (cons var vars-b)
                       (cons val vals-b)
                       vars-a)
                (loop1 (cdr l)
                       vars-b
                       vals-b
                       (cons var vars-a))))

            (let* ((result1
                     (let loop2 ((l vars-a))
                       (if (not (null? l))

                         (let* ((var (car l))
                                (val (val-of var)))
                           (new-seq source (env-declarations env)
                             (new-set source (env-declarations env) var val)
                             (loop2 (cdr l))))

                         (bind-in-order (cdr order)))))

                   (result2
                     (if (null? vars-b)
                       result1
                       (new-call* source (env-declarations env)
                         (new-prc source (env-declarations env)
                           #f
                           #f
                           vars-b
                           '()
                           #f
                           #f
                           result1)
                         vals-b)))

                   (result3
                     (if (null? vars-a)
                       result2
                       (new-call* source (env-declarations env)
                         (new-prc source (env-declarations env)
                           #f
                           #f
                           vars-a
                           '()
                           #f
                           #f
                           result2)
                         (map (lambda (var)
                                (new-cst
                                  source
                                  (env-declarations env)
                                  void-object))
                              vars-a)))))

              result3))))))

  (set-prc-names! vars vals)

  (bind-in-order
    (topological-sort
      (transitive-closure
        (dependency-graph vars vals)))))

(define (pt-begin source env use)
  (pt-sequence source (cdr (source-code source)) env use))

(define (pt-do source env use)
  (let* ((code (source-code source))
         (loop (new-temp-variable source 'do-temp))
         (bindings (map source-code (source-code (cadr code))))
         (vars (new-variables (map car bindings)))
         (init (map (lambda (x) (pt (cadr x) env 'true)) bindings))
         (env  (env-frame env vars))
         (step (map (lambda (x)
                      (pt (if (length? x 2) (car x) (caddr x)) env 'true))
                    bindings))
         (exit (source-code (caddr code))))
    (set-prc-names! vars init)
    (new-call* source (env-declarations env)
      (new-prc source (env-declarations env) #f #f (list loop) '() #f #f
        (new-call* source (env-declarations env)
          (new-ref source (env-declarations env) loop) init))
      (list
        (new-prc source (env-declarations env) #f #f vars '() #f #f
          (new-tst source (env-declarations env)
            (pt (car exit) env 'pred)
            (if (length? exit 1)
              (new-cst (caddr code) (env-declarations env) void-object)
              (pt-sequence (caddr code) (cdr exit) env use))
            (if (length? code 3)
              (new-call* source (env-declarations env)
                (new-ref source (env-declarations env) loop)
                step)
              (new-seq source (env-declarations env)
                (pt-sequence source (cdddr code) env 'none)
                (new-call* source (env-declarations env)
                  (new-ref source (env-declarations env)
                    loop)
                  step)))))))))

(define (pt-combination source env use)
  (let* ((code (source-code source))
         (oper (pt (car code) env 'true))
         (decl (node-decl oper)))
    (new-call* source (env-declarations env)
      oper
      (map (lambda (x) (pt x env 'true)) (cdr code)))))

(define (pt-delay source env use)
  (let ((code (source-code source)))
    (new-call* source (add-not-safe (env-declarations env))
      (new-ref-extended-bindings source **make-placeholder-sym env)
      (list (new-prc source (env-declarations env) #f #f '() '() #f #f
              (pt (cadr code) env 'true))))))

(define (pt-future source env use)
  (let ((decl (env-declarations env))
        (code (source-code source)))
    (new-fut source decl
      (pt (cadr code) env 'true))))

; Expression identification predicates and syntax checking.

(define (self-eval-expr? source)
  (let ((code (source-code source)))
    (or (number? code)
        (string? code)
        (char? code)
        (keyword-object? code)
        (eq? code #t)
        (eq? code false-object)
        (eq? code end-of-file-object)
        (eq? code optional-object)
        (eq? code rest-object)
        (eq? code key-object)
        (eq? code script-object))))

(define (quote-expr? source)
  (match quote-sym 1 source))

(define (quasiquote-expr? source)
  (match quasiquote-sym 1 source))

(define (unquote-expr? source)
  (match unquote-sym 1 source))

(define (unquote-splicing-expr? source)
  (match unquote-splicing-sym 1 source))

(define (var-expr? source env)
  (let ((code (source-code source)))
    (and (symbol-object? code)
         (not-reserved-identifier source env code)
         (not-macro source env code))))

(define (not-macro source env name)
  (if (env-lookup-macro env name)
    (pt-syntax-error source "Macro name can't be used as a variable:" name)
    #t))

(define (bindable-var? source env)
  (let ((code (source-code source)))
    (and (symbol-object? code)
         (not-reserved-identifier source env code))))

(define (not-reserved-identifier source env name)
  (if (or (memq name common-reserved-identifiers)
          (memq name (dialect-specific-reserved-identifiers
                       (scheme-dialect (env-declarations env)))))
    (pt-syntax-error
      source
      "Reserved identifier can't be used as a variable:" name)
    #t))

(define (set!-expr? source env)
  (and (match set!-sym 2 source)
       (var-expr? (cadr (source-code source)) env)))

(define (lambda-expr? source env)
  (match lambda-sym -2 source))

(define (if-expr? source)
  (and (match if-sym -2 source)
       (or (<= (length (source-code source)) 4)
           (pt-syntax-error source "Ill-formed special form" if-sym))))

(define (cond-expr? source)
  (and (match cond-sym -1 source)
       (proper-clauses? source)))

(define (and-expr? source)
  (match and-sym 0 source))

(define (or-expr? source)
  (match or-sym 0 source))

(define (case-expr? source)
  (and (match case-sym -2 source)
       (proper-case-clauses? source)))

(define (let-expr? source env)
  (and (match let-sym -2 source)
       (let ((code (source-code source)))
         (if (bindable-var? (cadr code) env)
           (and (proper-bindings? (caddr code) #t env)
                (or (> (length code) 3)
                    (pt-syntax-error source "Ill-formed named 'let'")))
           (proper-bindings? (cadr code) #t env)))))

(define (let*-expr? source env)
  (and (match let*-sym -2 source)
       (proper-bindings? (cadr (source-code source)) #f env)))

(define (letrec-expr? source env)
  (and (match letrec-sym -2 source)
       (proper-bindings? (cadr (source-code source)) #t env)))

(define (begin-expr? source)
  (match begin-sym -1 source))

(define (do-expr? source env)
  (and (match do-sym -2 source)
       (proper-do-bindings? source env)
       (proper-do-exit? source)))

(define (define-expr? source env)
  (and (match define-sym -1 source)
       (proper-definition? source env)
       (let ((v (definition-variable source)))
         (not-macro v env (source-code v)))))

(define (combination-expr? source)
  (let ((code (source-code source)))
    (and (pair? code)
         (let ((length (proper-length code)))
           (if length
             (or (> length 0)
                 (pt-syntax-error source "Ill-formed procedure call"))
             (pt-syntax-error source "Ill-terminated procedure call"))))))

(define (delay-expr? source env)
  (and (not (eq? (scheme-dialect (env-declarations env)) ieee-scheme-sym))
       (match delay-sym 1 source)))
       
(define (future-expr? source env)
  (and (eq? (scheme-dialect (env-declarations env)) multilisp-sym)
       (match future-sym 1 source)))
       
(define (macro-expr? source env)
  (let ((code (source-code source)))
    (and (pair? code)
         (symbol-object? (source-code (car code)))
         (let ((macr (env-lookup-macro env (source-code (car code)))))
           (and macr
                (let ((len (proper-length (cdr code))))
                  (if len
                    (let ((len* (+ len 1))
                          (size (car macr)))
                      (or (if (> size 0) (= len* size) (>= len* (- size)))
                          (pt-syntax-error source "Ill-formed macro form")))
                    (pt-syntax-error source "Ill-terminated macro form"))))))))

(define (define-macro-expr? source env)
  (and (match **define-macro-sym -1 source)
       (proper-definition? source env)))

(define (declare-expr? source)
  (match **declare-sym -1 source))

(define (include-expr? source)
  (match **include-sym 1 source))

(define (begin-defs-expr? source)
  (match begin-sym 0 source))

(define (match head size source)
  (let ((code (source-code source)))
    (and (pair? code)
         (eq? (source-code (car code)) head)
         (let ((length (proper-length (cdr code))))
           (if length
             (or (if (> size 0) (= length size) (>= length (- size)))
                 (pt-syntax-error source "Ill-formed special form" head))
             (pt-syntax-error source "Ill-terminated special form" head))))))

(define (proper-length l)
  (define (length l n)
    (cond ((pair? l) (length (cdr l) (+ n 1)))
          ((null? l) n)
          (else      #f)))
  (length l 0))

(define (proper-definition? source env)
  (let* ((code (source-code source))
         (pattern-source (cadr code))
         (pattern (source-code pattern-source))
         (body (cddr code)))
    (cond ((bindable-var? pattern-source env)
           (cond ((length? body 0) #t) ; an unbound variable
                 ((length? body 1) #t) ; a bound variable
                 (else
                  (pt-syntax-error source "Ill-formed definition body"))))
          ((pair? pattern)
           (if (length? body 0)
             (pt-syntax-error
              source
              "Body of a definition must have at least one expression"))
           (if (not (bindable-var? (car pattern) env))
             (pt-syntax-error
               (car pattern)
               "Procedure name must be an identifier")))
          (else
           (pt-syntax-error pattern-source "Ill-formed definition pattern")))))

(define (definition-variable def)
  (let* ((code (source-code def))
         (pattern (cadr code)))
    (if (pair? (source-code pattern))
      (car (source-code pattern))
      pattern)))

(define (definition-value def)
  (let ((code (source-code def))
        (loc (source-locat def)))
    (cond ((pair? (source-code (cadr code)))
           (make-source
             (cons (make-source lambda-sym loc)
                   (cons (parms->source (cdr (source-code (cadr code))) loc)
                         (cddr code)))
             loc))
          ((null? (cddr code))
           (make-source
             (list (make-source quote-sym loc) (make-source void-object loc))
             loc))
          (else
           (caddr code)))))

(define (parms->source parms loc)
  (if (or (pair? parms) (null? parms))
    (make-source parms loc)
    parms))

(define (proper-clauses? source)

  (define (proper-clauses clauses)
    (or (null? clauses)
        (let* ((clause-source (car clauses))
               (clause (source-code clause-source))
               (length (proper-length clause)))
          (if length
            (if (>= length 1)
              (if (eq? (source-code (car clause)) else-sym)
                (cond ((= length 1)
                       (pt-syntax-error
                         clause-source
                         "Else clause must have a body"))
                      ((not (null? (cdr clauses)))
                       (pt-syntax-error
                         clause-source
                         "Else clause must be the last clause"))
                      (else
                       (proper-clauses (cdr clauses))))
                (if (and (>= length 2)
                         (eq? (source-code (cadr clause)) =>-sym)
                         (not (= length 3)))
                  (pt-syntax-error
                    (cadr clause)
                    "'=>' must be followed by a single expression")
                  (proper-clauses (cdr clauses))))
              (pt-syntax-error clause-source "Ill-formed 'cond' clause"))
            (pt-syntax-error clause-source "Ill-terminated 'cond' clause")))))

  (proper-clauses (cdr (source-code source))))

(define (proper-case-clauses? source)

  (define (proper-case-clauses clauses)
    (or (null? clauses)
        (let* ((clause-source (car clauses))
               (clause (source-code clause-source))
               (length (proper-length clause)))
          (if length
            (if (>= length 2)
              (if (eq? (source-code (car clause)) else-sym)
                (if (not (null? (cdr clauses)))
                  (pt-syntax-error
                    clause-source
                    "Else clause must be the last clause")
                  (proper-case-clauses (cdr clauses)))
                (begin
                  (proper-selector-list? (car clause))
                  (proper-case-clauses (cdr clauses))))
              (pt-syntax-error
                clause-source
                "A 'case' clause must have a selector list and a body"))
            (pt-syntax-error clause-source "Ill-terminated 'case' clause")))))

  (proper-case-clauses (cddr (source-code source))))

(define (proper-selector-list? source)
  (let* ((code (source-code source))
         (length (proper-length code)))
    (if length
      (or (>= length 1)
          (pt-syntax-error
            source
            "Selector list must contain at least one element"))
      (pt-syntax-error source "Ill-terminated selector list"))))

(define (proper-bindings? bindings check-dupl? env)

  (define (proper-bindings l seen)
    (cond ((pair? l)
           (let* ((binding-source (car l))
                  (binding (source-code binding-source)))
             (if (eqv? (proper-length binding) 2)
               (let ((var (car binding)))
                 (if (bindable-var? var env)
                   (if (and check-dupl? (memq (source-code var) seen))
                     (pt-syntax-error var "Duplicate variable in bindings")
                     (proper-bindings (cdr l)
                                      (cons (source-code var) seen)))
                   (pt-syntax-error
                     var
                     "Binding variable must be an identifier")))
               (pt-syntax-error binding-source "Ill-formed binding"))))
          ((null? l)
           #t)
          (else
           (pt-syntax-error bindings "Ill-terminated binding list"))))
          
   (proper-bindings (source-code bindings) '()))

(define (proper-do-bindings? source env)
  (let ((bindings (cadr (source-code source))))

    (define (proper-bindings l seen)
      (cond ((pair? l)
             (let* ((binding-source (car l))
                    (binding (source-code binding-source))
                    (length (proper-length binding)))
               (if (or (eqv? length 2) (eqv? length 3))
                 (let ((var (car binding)))
                   (if (bindable-var? var env)
                     (if (memq (source-code var) seen)
                       (pt-syntax-error var "Duplicate variable in bindings")
                       (proper-bindings (cdr l)
                                        (cons (source-code var) seen)))
                     (pt-syntax-error
                       var
                       "Binding variable must be an identifier")))
                 (pt-syntax-error binding-source "Ill-formed binding"))))
            ((null? l)
             #t)
            (else
             (pt-syntax-error bindings "Ill-terminated binding list"))))

     (proper-bindings (source-code bindings) '())))

(define (proper-do-exit? source)
  (let* ((code (source-code (caddr (source-code source))))
         (length (proper-length code)))
    (if length
      (or (> length 0)
          (pt-syntax-error source "Ill-formed exit clause"))
      (pt-syntax-error source "Ill-terminated exit clause"))))

(define (include-filename source)
  (source-code (cadr (source-code source))))

(define (begin-defs-body source)
  (cdr (source-code source)))

(define (length? l n)
  (cond ((null? l) (= n 0))
        ((> n 0)   (length? (cdr l) (- n 1)))
        (else      #f)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Declaration handling:
; --------------------

; A declaration has the form: (##declare <item1> <item2> ...)
;
; an <item> can be one of 6 types:
;
; - flag declaration           : (<id>)
; - parameterized declaration  : (<id> <parameter>)
; - boolean declaration        : (<id>)  or  (NOT <id>)
; - namable declaration        : (<id> <name>...)
; - namable boolean declaration: (<id> <name>...)  or  (NOT <id> <name>...)
; - namable string declaration : (<id> <string> <name>...)

(define (transform-declaration source)
  (let ((code (source-code source)))
    (if (not (pair? code))
      (pt-syntax-error source "Ill-formed declaration")
      (let* ((pos (not (eq? (source-code (car code)) not-sym)))
             (x (if pos code (cdr code))))
        (if (not (pair? x))
          (pt-syntax-error source "Ill-formed declaration")
          (let* ((id-source (car x))
                 (id (source-code id-source)))

            (cond ((not (symbol-object? id))
                   (pt-syntax-error
                     id-source
                     "Declaration name must be an identifier"))

                  ((assq id flag-declarations)
                   (cond ((not pos)
                          (pt-syntax-error
                            id-source
                            "Declaration can't be negated"))
                         ((null? (cdr x))
                          (flag-decl
                            source
                            (cdr (assq id flag-declarations))
                            id))
                         (else
                          (pt-syntax-error source "Ill-formed declaration"))))

                  ((memq id parameterized-declarations)
                   (cond ((not pos)
                          (pt-syntax-error
                            id-source
                            "Declaration can't be negated"))
                         ((eqv? (proper-length x) 2)
                          (let ((parm (source->expression (cadr x))))
                            (if (not (and (integer? parm) (exact? parm)))
                              (pt-syntax-error source "Exact integer expected")
                              (parameterized-decl source id parm))))
                         (else
                          (pt-syntax-error source "Ill-formed declaration"))))

                  ((memq id boolean-declarations)
                   (if (null? (cdr x))
                     (boolean-decl source id pos)
                     (pt-syntax-error source "Ill-formed declaration")))

                  ((assq id namable-declarations)
                   (cond ((not pos)
                          (pt-syntax-error
                            id-source
                            "Declaration can't be negated"))
                         (else
                          (namable-decl
                            source
                            (cdr (assq id namable-declarations))
                            id
                            (map source->expression (cdr x))))))

                  ((memq id namable-boolean-declarations)
                   (namable-boolean-decl
                     source
                     id
                     pos
                     (map source->expression (cdr x))))

                  ((memq id namable-string-declarations)
                   (if (not (pair? (cdr x)))
                     (pt-syntax-error source "Ill-formed declaration")
                     (let* ((str-source (cadr x))
                            (str (source-code str-source)))
                       (cond ((not pos)
                              (pt-syntax-error
                                id-source
                                "Declaration can't be negated"))
                             ((not (string? str))
                              (pt-syntax-error str-source "String expected"))
                             (else
                              (namable-string-decl
                                source
                                id
                                str
                                (map source->expression (cddr x))))))))

                  (else
                   (pt-syntax-error id-source "Unknown declaration")))))))))

(define (add-declarations source env)
  (let loop ((l (cdr (source-code source))) (env env))
    (if (pair? l)
      (loop (cdr l) (env-declare env (transform-declaration (car l))))
      env)))

(define (add-decl d decl)
  (env-declare decl d))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Macro handling:
; --------------

(define (add-macro source env)

  (define (form-size parms)
    (let loop ((l parms) (n 1))
      (if (pair? l)
        (loop (cdr l) (+ n 1))
        (if (null? l) n (- n)))))

  (define (error-proc . msgs)
    (apply compiler-user-error
           (cons (source-locat source)
                 (cons "(in macro body)" msgs))))

  (let ((var (definition-variable source))
        (proc (definition-value source)))
    (if (lambda-expr? proc env)
      (env-macro env
                 (source-code var)
                 (cons (form-size (source->parms (cadr (source-code proc))))
                       (scheme-global-eval (source->expression proc)
                                           error-proc)))
      (pt-syntax-error source "Macro value must be a lambda expression"))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (ptree.begin! info-port) ; initialize module
  (set! *ptree-port* info-port)
  '())

(define (ptree.end!) ; finalize module
  '())

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; Stuff local to the module:

(define *ptree-port* '())

;==============================================================================
