;==============================================================================

; file: "_source.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(include "fixnum.scm")

;------------------------------------------------------------------------------
;
; Source code manipulation module:
; -------------------------------

; This module contains procedures to manipulate source code representations
; read in from Scheme source files.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; 'source file' manipulation

(define (open-error filename)
  (compiler-error "Can't find file" filename))

(define (open-sf filename)
  (if (equal? (path-extension filename) "")

    (let loop ((exts scm-file-exts))
      (if (pair? exts)
        (let* ((ext (car exts))
               (full-name (string-append filename ext))
               (port (open-input-file* full-name)))
          (if port
            (vector port full-name 0 1 0)
            (loop (cdr exts))))
        (open-error filename)))

    (let ((port (open-input-file* filename)))
      (if port
        (vector port filename 0 1 0)
        (open-error filename)))))

(define (close-sf sf)
  (close-input-port (vector-ref sf 0)))

(define (sf-read-char sf)
  (let ((c (read-char (vector-ref sf 0))))
    (cond ((eof-object? c)
           end-of-file-object)
          ((char=? c #\newline)
           (vector-set! sf 3 (+ (vector-ref sf 3) 1))
           (vector-set! sf 4 0)
           c)
          (else
           (vector-set! sf 4 (+ (vector-ref sf 4) 1))
           c))))

(define (sf-peek-char sf)
  (let ((c (peek-char (vector-ref sf 0))))
    (cond ((eof-object? c)
           end-of-file-object)
          (else
           c))))

(define (sf-read-error sf msg . args)
  (apply compiler-user-error
         (cons (sf->locat sf)
               (cons (string-append "Read error -- " msg) args))))

(define (sf-read-error-on source msg . args)
  (apply compiler-user-error
         (cons (source-locat source)
               (cons (string-append "Read error -- " msg) args))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; 'location' manipulation

(define (sf->locat sf)
  (vector 'file
          (vector-ref sf 1)
          (vector-ref sf 2)
          (vector-ref sf 3)
          (vector-ref sf 4)))

(define (expr->locat expr source)
  (vector 'expr
          expr
          source))

(define (locat-show loc)
  (if loc

    (case (vector-ref loc 0)
      ((file)
       (if (pinpoint-error
             (vector-ref loc 1)
             (vector-ref loc 3)
             (vector-ref loc 4))
         (begin
           (display "file \"")
           (display (vector-ref loc 1))
           (display "\", line ")
           (display (vector-ref loc 3))
           (display ", character ")
           (display (vector-ref loc 4)))))
      ((expr)
       (display "expression ")
       (write (vector-ref loc 1))
       (if (vector-ref loc 2)
         (begin
           (display " ")
           (locat-show (source-locat (vector-ref loc 2))))))
      (else
       (compiler-internal-error "locat-show, unknown location tag")))

    (display "unknown location")))

(define (locat-filename loc)
  (if loc
    (case (vector-ref loc 0)
      ((file)
       (vector-ref loc 1))
      ((expr)
       (let ((source (vector-ref loc 2)))
         (if source
           (locat-filename (source-locat source))
           "")))
      (else
       (compiler-internal-error "locat-filename, unknown location tag")))
    ""))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;
; 'source' manipulation

(define (make-source code locat)
  (vector code locat))

(define (source-code x)        (vector-ref x 0))
(define (source-code-set! x y) (vector-set! x 0 y) x)
(define (source-locat x)       (vector-ref x 1))

; (expression->source expr source) returns the source that represent the Scheme
; expression 'expr' and is related to the source 'source' (#f if no relation).

(define (expression->source expr source)

  (define (expr->source x)
    (make-source (cond ((pair? x)
                        (list-convert x))
                       ((vector? x)
                        (vector-convert x))
                       (else
                        x))
                 (expr->locat x source)))

  (define (list-convert l)
    (cons (expr->source (car l))
          (list-tail-convert (cdr l))))

  (define (list-tail-convert l)
    (cond ((pair? l)
           (if (quoting-form? l) ; so that macros which generate quoting-forms
             (expr->source l)    ; at the tail of a list work properly
             (cons (expr->source (car l))
                   (list-tail-convert (cdr l)))))
          ((null? l)
           '())
          (else
           (expr->source l))))

  (define (quoting-form? x)
    (let ((first (car x))
          (rest (cdr x)))
      (and (pair? rest)
           (null? (cdr rest))
           (or (eq? first quote-sym)
               (eq? first quasiquote-sym)
               (eq? first unquote-sym)
               (eq? first unquote-splicing-sym)))))

  (define (vector-convert v)
    (let* ((len (vector-length v))
           (x (make-vector len)))
      (let loop ((i (- len 1)))
        (if (>= i 0)
          (begin
            (vector-set! x i (expr->source (vector-ref v i)))
            (loop (- i 1)))))
      x))

  (expr->source expr))

; (source->expression source) returns the Scheme expression represented by the
; source 'source'.  Note that every call with the same argument returns a
; different (i.e. non eq?) expression.

(define (source->expression source)

  (define (list->expression l)
    (cond ((pair? l)
           (cons (source->expression (car l)) (list->expression (cdr l))))
          ((null? l)
           '())
          (else
           (source->expression l))))

  (define (vector->expression v)
    (let* ((len (vector-length v))
           (x (make-vector len)))
      (let loop ((i (- len 1)))
        (if (>= i 0)
          (begin
            (vector-set! x i (source->expression (vector-ref v i)))
            (loop (- i 1)))))
      x))

  (let ((code (source-code source)))
    (cond ((pair? code)   (list->expression code))
          ((vector? code) (vector->expression code))
          (else           code))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; (file->sources filename info-port) returns a list of the source
; representation for each of the expressions contained in the file 'filename'.

(define (file->sources filename info-port)
  (let ((expanded-filename (path-expand filename)))
    (if (not expanded-filename)
      (open-error filename)
      (let ((sf (open-sf expanded-filename)))

        (define (read-sources) ; return list of all sources in file
          (let ((source (read-source sf)))
            (if (not (end-of-file-object? source))
              (begin
                (if info-port (display "." info-port))
                (cons source (read-sources)))
              '())))

        (if info-port
          (begin
            (display "(reading \"" info-port)
            (display (locat-filename (sf->locat sf)) info-port)
            (display "\"" info-port)))

        (let ((sources (read-sources)))

          (if info-port (display ")" info-port))

          (close-sf sf)

          sources)))))

(define (file->sources* filename info-port loc)
  (file->sources
    (if (path-absolute? filename)
      filename
      (string-append
        (path-directory (locat-filename loc))
        filename))
    info-port))

; To implement character classes, 'char-indexed-vectors' are used.
; This data structure behaves like normal vectors except that they
; are indexed using characters.

(define direct-access-size 128) ; first 128 chars are indexed directly

(define (make-char-indexed-vector init)
  (let ((tbl (make-vector (+ direct-access-size 2) init)))
    (vector-set! tbl direct-access-size '())
    tbl))

(define (char-indexed-vector-ref tbl c)
  (let ((i (character->unicode c)))
    (if (and (>= i 0) (< i direct-access-size))
      (vector-ref tbl i)
      (let ((x (assq i (vector-ref tbl direct-access-size))))
        (if x
          (cdr x)
          (vector-ref tbl (+ direct-access-size 1)))))))

(define (char-indexed-vector-set! tbl c val)
  (let ((i (character->unicode c)))
    (if (and (>= i 0) (< i direct-access-size))
      (vector-set! tbl i val)
      (let ((x (assq i (vector-ref tbl direct-access-size))))
        (if x
          (set-cdr! x val)
          (vector-set!
            tbl
            direct-access-size
            (cons (cons i val) (vector-ref tbl direct-access-size))))))))

; (read-source sf) returns the source for the next expression in the source
; file 'sf'.

(define (read-source sf)

  (define (read-char*)
    (let ((c (sf-read-char sf)))
      (if (end-of-file-object? c)
        (sf-read-error sf "Premature end of file encountered")
        c)))

  (define (read-non-whitespace-char)
    (let ((c (read-char*)))
      (cond ((< 0 (char-indexed-vector-ref read-table c))
             (read-non-whitespace-char))
            ((char=? c #\;)
             (let loop ()
               (if (not (char=? (read-char*) #\newline))
                 (loop)
                 (read-non-whitespace-char))))
            (else
             c))))

  (define (delimiter? c)
    (or (end-of-file-object? c)
        (not (= (char-indexed-vector-ref read-table c) 0))))

  (define (read-list first)
    (let ((result (cons first '())))
      (let loop ((end result))
        (let ((c (read-non-whitespace-char)))
          (cond ((char=? c #\)))
                ((and (char=? c #\.) (delimiter? (sf-peek-char sf)))
                 (let ((x (read-source sf)))
                   (if (char=? (read-non-whitespace-char) #\))
                     (set-cdr! end x)
                     (sf-read-error sf "')' expected"))))
                (else
                 (let ((tail (cons (rd* c) '())))
                   (set-cdr! end tail)
                   (loop tail))))))
      result))

  (define (read-blist first)
    (let ((result (cons first '())))
      (let loop ((end result))
        (let ((c (read-non-whitespace-char)))
          (cond ((char=? c #\]))
                ((and (char=? c #\.) (delimiter? (sf-peek-char sf)))
                 (let ((x (read-source sf)))
                   (if (char=? (read-non-whitespace-char) #\])
                     (set-cdr! end x)
                     (sf-read-error sf "']' expected"))))
                (else
                 (let ((tail (cons (rd* c) '())))
                   (set-cdr! end tail)
                   (loop tail))))))
      result))

  (define (read-vector kind)

    (define (loop i)
      (let ((c (read-non-whitespace-char)))
        (if (char=? c #\))
          (case kind
            ((vector)    (make-vector i))
            ((u8vector)  (make-u8vect i))
            ((u16vector) (make-u16vect i))
            ((u32vector) (make-u32vect i))
            ((f32vector) (make-f32vect i))
            ((f64vector) (make-f64vect i)))
          (let* ((x (rd* c))
                 (v (loop (+ i 1))))
            (case kind
              ((vector)
               (vector-set! v i x))
              ((u8vector)
               (let ((n (source-code x)))
                 (if (not (exact-integer-check n -128 255))
                   (sf-read-error-on x "8 bit exact integer expected"))
                 (u8vect-set! v i n)))
              ((u16vector)
               (let ((n (source-code x)))
                 (if (not (exact-integer-check n -32768 65535))
                   (sf-read-error-on x "16 bit exact integer expected"))
                 (u16vect-set! v i n)))
              ((u32vector)
               (let ((n (source-code x)))
                 (if (not (exact-integer-check n -2147483648 4294967295))
                   (sf-read-error-on x "32 bit exact integer expected"))
                 (u32vect-set! v i n)))
              ((f32vector)
               (let ((n (source-code x)))
                 (if (not (inexact-real-check n))
                   (sf-read-error-on x "inexact real expected"))
                 (f32vect-set! v i n)))
              ((f64vector)
               (let ((n (source-code x)))
                 (if (not (inexact-real-check n))
                   (sf-read-error-on x "inexact real expected"))
                 (f64vect-set! v i n))))
            v))))

    (loop 0))

  (define (read-vector* kind)
    (let ((c (read-char*)))
      (if (char=? c #\()
        (read-vector kind)
        (sf-read-error sf "'(' expected"))))

  (define (exact-integer-check n lo hi)
    (and (integer? n)
         (exact? n)
         (in-integer-range? n lo hi)))

  (define (inexact-real-check n)
    (and (real? n)
         (not (exact? n))))

  (define (char-octal? c)
    (and (not (char<? c #\0)) (not (char<? #\7 c))))

  (define (char-hexadecimal? c)
    (or (and (not (char<? c #\0)) (not (char<? #\9 c)))
        (and (not (char<? c #\a)) (not (char<? #\f c)))
        (and (not (char<? c #\A)) (not (char<? #\F c)))))

  (define (unicode n)
    (if (in-unicode-range? n)
      (unicode->character n)
      (sf-read-error sf "Character encoding out of range")))

  (define (read-escape-octal c)
    (let ((str (let loop ((i 1))
                 (let ((c (sf-peek-char sf)))
                   (if (and (< i 3)
                            (not (end-of-file-object? c))
                            (char-octal? c))
                     (begin
                       (sf-read-char sf)
                       (let ((s (loop (+ i 1))))
                         (string-set! s i c)
                         s))
                     (make-string i #\space))))))
      (string-set! str 0 c)
      (unicode (string->number str 8))))

  (define (read-escape-hexadecimal)
    (let ((c (sf-peek-char sf)))
      (if (or (end-of-file-object? c)
              (not (char-hexadecimal? c)))
        (sf-read-error sf "\\x not followed by a hexadecimal digit")
        (begin
          (sf-read-char sf)
          (let ((str (let loop ((i 1))
                       (let ((c (sf-peek-char sf)))
                         (if (and (not (end-of-file-object? c))
                                  (char-hexadecimal? c))
                           (begin
                             (sf-read-char sf)
                             (let ((s (loop (+ i 1))))
                               (string-set! s i c)
                               s))
                           (make-string i #\space))))))
            (string-set! str 0 c)
            (unicode (string->number str 16)))))))

  (define (read-escape special-escape)
    (let ((c (read-char*)))
      (cond ((char-octal? c)
             (read-escape-octal c))
            ((char=? c #\x)
             (read-escape-hexadecimal))
            ((char=? c special-escape)
             special-escape)
            (else
             (let ((x (assq c escaped-char-table)))
               (if x
                 (cdr x)
                 (sf-read-error sf "Unknown escaped character:" c)))))))

  (define (read-string closing-char)
    (define (loop i)
      (let ((c (read-char*)))
        (cond ((char=? c closing-char)
               (make-string i #\space))
              ((char=? c #\\)
               (let* ((c (read-escape closing-char))
                      (s (loop (+ i 1))))
                 (string-set! s i c)
                 s))
              (else
               (let ((s (loop (+ i 1))))
                 (string-set! s i c)
                 s)))))
    (loop 0))

  (define (read-symbol/keyword/number-string i)
    (if (delimiter? (sf-peek-char sf))
      (make-string i #\space)
      (let* ((c (sf-read-char sf))
             (s (read-symbol/keyword/number-string (+ i 1))))
        (string-set! s i (convert-case c))
        s)))

  (define (read-symbol/keyword/number c)
    (let ((s (read-symbol/keyword/number-string 1)))
      (string-set! s 0 (convert-case c))
      (or (string->number s 10)
          (parse-keyword s)
          (string->symbol s))))

  (define (read-sharp-token c)
    (let ((s (read-symbol/keyword/number-string 2)))
      (string-set! s 0 #\#)
      (string-set! s 1 (convert-case c))
      (or (string->number s 10)
          (cond ((string=? s "#f")
                 false-object)
                ((string=? s "#t")
                 #t)
                ((string=? s "#u8")
                 (read-vector* 'u8vector))
                ((string=? s "#u16")
                 (read-vector* 'u16vector))
                ((string=? s "#u32")
                 (read-vector* 'u32vector))
                ((string=? s "#f32")
                 (read-vector* 'f32vector))
                ((string=? s "#f64")
                 (read-vector* 'f64vector))
                (else
                 (sf-read-error sf "Unknown '#' token:" s))))))

  (define (read-sharp-sharp-symbol)
    (let ((s (read-symbol/keyword/number-string 2)))
      (string-set! s 0 #\#)
      (string-set! s 1 #\#)
      (string->symbol s)))

  (define (read-sharp-bang-token)
    (let ((s (read-symbol/keyword/number-string 0)))
      (string->symbol s)))

  (define (rd)
    (let ((c (sf-read-char sf)))
      (cond ((end-of-file-object? c) ;************fix
             c)
            ((< 0 (char-indexed-vector-ref read-table c))
             (rd))
            ((char=? c #\;)
             (let loop ()
               (let ((c (sf-read-char sf)))
                 (cond ((end-of-file-object? c)
                        c)
                       ((char=? c #\newline)
                        (rd))
                       (else
                        (loop))))))
            (else
             (rd* c)))))

  (define (rd* c)
    (let ((source (make-source #f (sf->locat sf))))
      (source-code-set!
        source
        (cond ((char=? c #\()
               (let ((x (read-non-whitespace-char)))
                 (if (char=? x #\))
                   '()
                   (read-list (rd* x)))))
              ((char=? c #\[)
               (let ((x (read-non-whitespace-char)))
                 (if (char=? x #\])
                   '()
                   (read-blist (rd* x)))))
              ((char=? c #\#)
               (let ((c (sf-read-char sf)))
                 (cond ((char=? c #\()
                        (read-vector 'vector))
                       ((char=? c #\\)
                        (let ((c (read-char*)))
                          (if (or (not (char-alphabetic? c))
                                  (delimiter? (sf-peek-char sf)))
                            c
                            (let ((name (read-symbol/keyword/number c)))
                              (let ((x (assq name named-char-table)))
                                (if x
                                  (cdr x)
                                  (sf-read-error sf "Unknown character name:" name)))))))
                       ((char=? c #\!)
                        (let ((name (read-sharp-bang-token)))
                          (let ((x (assq name sharp-bang-table)))
                            (if x
                              (cdr x)
                              (sf-read-error sf "Unknown '#!' token:" name)))))
                       ((char=? c #\#)
                        (read-sharp-sharp-symbol))
                       (else
                        (read-sharp-token c)))))
              ((char=? c #\")
               (read-string #\"))
              ((char=? c #\|)
               (string->symbol (read-string #\|)))
              ((char=? c #\')
               (list (make-source quote-sym (sf->locat sf))
                     (read-source sf)))
              ((char=? c #\`)
               (list (make-source quasiquote-sym (sf->locat sf))
                     (read-source sf)))
              ((char=? c #\,)
               (if (eq? (sf-peek-char sf) #\@)
                 (let ((x (make-source unquote-splicing-sym (sf->locat sf))))
                   (sf-read-char sf)
                   (list x (read-source sf)))
                 (list (make-source unquote-sym (sf->locat sf))
                       (read-source sf))))
              ((char=? c #\))
               (sf-read-error sf "Misplaced ')'"))
              ((char=? c #\])
               (sf-read-error sf "Misplaced ']'"))
              ((or (char=? c #\{)
                   (char=? c #\}))
               (sf-read-error sf "Illegal character" c))
              (else
               (if (char=? c #\.)
                 (if (delimiter? (sf-peek-char sf))
                   (sf-read-error sf "Misplaced '.'")))
               (read-symbol/keyword/number c))))))

  (rd))

(define named-char-table
  (list
    (cons (string->canonical-symbol "newline")   #\newline)
    (cons (string->canonical-symbol "nul")       (unicode->character 0))
    (cons (string->canonical-symbol "backspace") (unicode->character 8))
    (cons (string->canonical-symbol "tab")       (unicode->character 9))
    (cons (string->canonical-symbol "linefeed")  (unicode->character 10))
    (cons (string->canonical-symbol "page")      (unicode->character 12))
    (cons (string->canonical-symbol "return")    (unicode->character 13))
    (cons (string->canonical-symbol "space")     #\ )
    (cons (string->canonical-symbol "rubout")    (unicode->character 127))))

(define escaped-char-table
  (list
    (cons #\a (unicode->character 7))
    (cons #\b (unicode->character 8))
    (cons #\t (unicode->character 9))
    (cons #\n #\newline)
    (cons #\v (unicode->character 11))
    (cons #\f (unicode->character 12))
    (cons #\r (unicode->character 13))
    (cons #\\ #\\)))

(define sharp-bang-table
  (list
    (cons (string->canonical-symbol "eof")      end-of-file-object)
    (cons (string->canonical-symbol "optional") optional-object)
    (cons (string->canonical-symbol "rest")     rest-object)
    (cons (string->canonical-symbol "key")      key-object)
    (cons (string->canonical-symbol "")         script-object)))

(define read-table
  (let ((rt (make-char-indexed-vector 0)))

    ; setup whitespace chars

    (char-indexed-vector-set! rt (unicode->character 9)  1)
    (char-indexed-vector-set! rt #\newline               1)
    (char-indexed-vector-set! rt (unicode->character 12) 1)
    (char-indexed-vector-set! rt (unicode->character 13) 1)
    (char-indexed-vector-set! rt #\                      1)

    ; setup other delimiters

    (char-indexed-vector-set! rt #\; -1)
    (char-indexed-vector-set! rt #\( -1)
    (char-indexed-vector-set! rt #\) -1)
    (char-indexed-vector-set! rt #\[ -1)
    (char-indexed-vector-set! rt #\] -1)
    (char-indexed-vector-set! rt #\{ -1)
    (char-indexed-vector-set! rt #\} -1)
    (char-indexed-vector-set! rt #\" -1)
    (char-indexed-vector-set! rt #\' -1)
    (char-indexed-vector-set! rt #\` -1)
    (char-indexed-vector-set! rt #\, -1)
    (char-indexed-vector-set! rt #\| -1)

    rt))

;==============================================================================
