;==============================================================================

; file: "_eval.scm"

; Copyright (C) 1994-1997 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

(define ##self-var     (##make-uninterned-symbol "<self>"))
(define ##selector-var (##make-uninterned-symbol "<selector>"))
(define ##do-loop-var  (##make-uninterned-symbol "<do-loop>"))

(define (##mk-code* code-prc cte lst n)
  (let ((code (##make-vector (##fixnum.+ (##length lst) (##fixnum.+ n 3)) #f)))
    (##vector-set! code 0 #f)
    (##vector-set! code 1 code-prc)
    (##vector-set! code 2 cte)
    (let loop ((i 0) (l lst))
      (if (##pair? l)
        (let ((child (##car l)))
          (##vector-set! child 0 code)
          (code-set! code i child)
          (loop (##fixnum.+ i 1) (##cdr l)))
        code))))

;==============================================================================

; Compiler

;------------------------------------------------------------------------------

; Compile time environments

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of local variables (up and over) and global variable.

(##define-macro (mk-loc-access up over) `(##cons ,up ,over))
(##define-macro (loc-access? x) `(##pair? ,x))
(##define-macro (loc-access-up x) `(##car ,x))
(##define-macro (loc-access-over x) `(##cdr ,x))

(##define-macro (mk-glo-access id)
  `(or (make-global-var ,id)
       (ct-error-global-env-overflow ,id)))

(##define-macro (glo-access? x)
  `(##not (##pair? ,x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Representation of compile time environments

; There are 4 types of structures in a compile time environment:
;
;    top        end of the environment and container for current state
;    frame      binding context for variables
;    macro      binding context for a macro
;    namespace  binding context for a namespace

(define (##cte-top top-cte)
  (##vector top-cte))

(define (##cte-top? cte)
  (##fixnum.= (##vector-length cte) 1))

(define (##cte-top-cte cte)
  (##vector-ref cte 0))

(define (##cte-top-cte-set! cte new-cte)
  (##vector-set! cte 0 new-cte))

(define (##cte-parent-cte cte)
  (##vector-ref cte 0))

(define (##cte-frame parent-cte vars)
  (##vector parent-cte vars))

(define (##cte-frame? cte)
  (##fixnum.= (##vector-length cte) 2))

(define (##cte-frame-vars cte)
  (##vector-ref cte 1))

(define (##cte-macro parent-cte name def)
  (##vector parent-cte name def))

(define (##cte-macro? cte)
  (and (##fixnum.= (##vector-length cte) 3)
       (##not (##string? (##vector-ref cte 1))))) ; distinguish from namespace

(define (##cte-macro-name cte)
  (##vector-ref cte 1))

(define (##cte-macro-def cte)
  (##vector-ref cte 2))

(define (##cte-namespace parent-cte prefix vars)
  (##vector parent-cte prefix vars))

(define (##cte-namespace? cte)
  (and (##fixnum.= (##vector-length cte) 3)
       (##string? (##vector-ref cte 1)))) ; distinguish from macro

(define (##cte-namespace-prefix cte)
  (##vector-ref cte 1))

(define (##cte-namespace-vars cte)
  (##vector-ref cte 2))

(define (##cte-relink cte new-parent-cte)
  (if new-parent-cte
    (cond ((##cte-frame? cte)
           (##cte-frame new-parent-cte
                        (##cte-frame-vars cte)))
          ((##cte-macro? cte)
           (##cte-macro new-parent-cte
                        (##cte-macro-name cte)
                        (##cte-macro-def cte)))
          ((##cte-namespace? cte)
           (##cte-namespace new-parent-cte
                            (##cte-namespace-prefix cte)
                            (##cte-namespace-vars cte))))
    #f))

(define (##cte-add-macro parent-cte name def)

  (define (replace cte)
    (cond ((##cte-top? cte)
           #f)
          ((and (##cte-macro? cte) (##eq? name (##cte-macro-name cte)))
           (##cte-macro (##cte-parent-cte cte) name def))
          ((##cte-namespace? cte)
           #f) ; don't go beyond a namespace declaration
          (else
           (##cte-relink cte (replace (##cte-parent-cte cte))))))

  (or (replace parent-cte)
      (##cte-macro parent-cte name def)))

(define (##cte-add-namespace parent-cte prefix vars)

  (define (replace cte)
    (cond ((##cte-top? cte)
           #f)
          ((##cte-namespace? cte)
           (if (##null? (##cte-namespace-vars cte))
             (##cte-namespace (##cte-parent-cte cte) prefix vars)
             (replace (##cte-parent-cte cte))))
          (else
           #f))) ; don't go beyond a frame or macro definition

  (if (##null? vars)
    (or (replace parent-cte)
        (##cte-namespace parent-cte prefix vars))
    (##cte-namespace parent-cte prefix vars)))

(define (##check-declarations src decls)

  (define (err) (ct-error-syntax "Ill-formed declaration"))

  (let loop ((decls decls))
    (if (##pair? decls)
      (let ((decl (##car decls)))
        (if (##pair? decl)
          (let ((head (##car decl)))
          (if (##eq? head 'namespace)
            (let ((rest (##cdr decl)))
              (if (##pair? rest)
                (let ((prefix (##car rest)))
                  (if (and (##string? prefix)
                           (##valid-prefix? prefix))
                    (##for-each (lambda (x)
                                  (if (##not (##symbol? x))
                                    (err)))
                                (##cdr rest))
                    (err)))
                (err))))))
        (loop (##cdr decls))))))

(define (##cte-add-declarations parent-cte src)
  (let ((decls (##cdr src)))
    (##check-declarations src decls)
    (let loop ((cte parent-cte) (decls decls))
      (if (##pair? decls)
        (let ((decl (##car decls)))
          (if (and (##pair? decl)
                   (##eq? (##car decl) 'namespace))
            (loop (##cte-add-namespace cte (##cadr decl) (##cddr decl))
                  (##cdr decls))
            (loop cte
                  (##cdr decls))))
        cte))))

(define (##cte-get-top-cte cte)
  (if (##cte-top? cte)
    cte
    (##cte-get-top-cte (##cte-parent-cte cte))))

(define (##cte-mutate-top-cte! cte proc)
  (let ((top-cte (##cte-get-top-cte cte)))
    (##cte-top-cte-set! top-cte (proc (##cte-top-cte top-cte)))))

(define (##make-top-cte)
  (let ((top-cte (##cte-top #f)))
    (##cte-top-cte-set! top-cte top-cte)
    top-cte))

(define (##top-cte-add-macro! top-cte name def)
  (let ((global-name (##cte-global-macro-name (##cte-top-cte top-cte) name)))
    (##cte-mutate-top-cte!
      top-cte
      (lambda (cte) (##cte-add-macro cte global-name def)))))

(define (##top-cte-add-declarations! top-cte src)
  (##cte-mutate-top-cte!
    top-cte
    (lambda (cte) (##cte-add-declarations cte src))))

(define (##top-cte-clone top-cte)
  (let ((new-top-cte (##cte-top #f)))

    (define (clone cte)
      (if (##cte-top? cte)
        new-top-cte
        (##cte-relink cte (clone (##cte-parent-cte cte)))))

    (##cte-top-cte-set! new-top-cte (clone (##cte-top-cte top-cte)))
    new-top-cte))

(define (##cte-lookup cte name)
  (##declare (inlining-limit 500)) ; inline CTE access procedures
  (let loop ((name name) (full? (##full-name? name)) (cte cte) (up 0))
    (if (##cte-top? cte)
      (##vector 'not-found name)
      (let ((parent-cte (##cte-parent-cte cte)))
        (cond ((##cte-frame? cte)
               (let* ((vars (##cte-frame-vars cte))
                      (x (##memq name vars)))
                 (if x
                   (##vector
                     'var
                     name
                     up
                     (##fixnum.+ (##fixnum.- (##length vars) (##length x)) 1))
                   (loop name full? parent-cte (##fixnum.+ up 1)))))
              ((##cte-macro? cte)
               (if (##eq? name (##cte-macro-name cte))
                 (##vector 'macro name (##cte-macro-def cte))
                 (loop name full? parent-cte up)))
              ((and (##not full?) (##cte-namespace? cte))
               (let ((vars (##cte-namespace-vars cte)))
                 (if (or (##null? vars) (##memq name vars))
                   (loop (##make-full-name (##cte-namespace-prefix cte) name)
                         #t
                         parent-cte
                         up)
                   (loop name full? parent-cte up))))
              (else
               (loop name full? parent-cte up)))))))

(define (##cte-global-macro-name cte name)
  (if (##full-name? name)
    name
    (let loop ((cte cte))
      (if (##cte-top? cte)
        name
        (let ((parent-cte (##cte-parent-cte cte)))
          (cond ((##cte-namespace? cte)
                 (let ((vars (##cte-namespace-vars cte)))
                   (if (or (##null? vars) (##memq name vars))
                     (##make-full-name (##cte-namespace-prefix cte) name)
                     (loop parent-cte))))
                (else
                 (loop parent-cte))))))))

(define (##full-name? sym) ; full name if it contains a "#"
  (let ((str (##symbol->string sym)))
    (let loop ((i (##fixnum.- (##string-length str) 1)))
      (if (##fixnum.< i 0)
        #f
        (if (##char=? (##string-ref str i) #\#)
          #t
          (loop (##fixnum.- i 1)))))))

(define (##make-full-name prefix sym)
  (if (##fixnum.= (##string-length prefix) 0)
    sym
    (##string->symbol (##string-append prefix (##symbol->string sym)))))

(define (##valid-prefix? str)      ; non-null name followed by a "#" at end is
  (let ((l (##string-length str))) ; valid as is the special prefix ""
    (or (##fixnum.= l 0)
        (and (##not (##fixnum.< l 2))
             (##char=? (##string-ref str (##fixnum.- l 1)) #\#)))))

(define (##var-lookup cte name src)
  (let ((ind (##cte-lookup cte name)))
    (case (##vector-ref ind 0)
      ((not-found)
       (mk-glo-access (##vector-ref ind 1)))
      ((var)
       (mk-loc-access (##vector-ref ind 2) (##vector-ref ind 3)))
      (else
       (ct-error-syntax "Macro name can't be used as a variable:" name)))))

(define ##macro-lookup #f)
(set! ##macro-lookup
  (lambda (cte name)
    (and (##symbol? name)
         (let ((ind (##cte-lookup cte name)))
           (case (##vector-ref ind 0)
             ((macro)
              (##vector-ref ind 2))
             (else
              #f))))))

(define ##macro-expand #f)
(set! ##macro-expand
  (lambda (cte src def)
    (##shape src src (##car def))
    (##apply (##cdr def) (##cdr src))))

(define ##macro-def #f)
(set! ##macro-def
  (lambda (src)
    (let ((src (##force-list src)))
      (if (and (##pair? src) (##eq? (##car src) 'lambda))
        (begin
          (##shape src src -3)
          (##cons (##form-size (##force-list (##cadr src)))
                  (##eval-top src ##interaction-cte)))
        (ct-error-syntax "Macro value must be a lambda expression")))))

(define (##form-size parms)
  (let loop ((l parms) (n 1))
    (if (##pair? l)
      (loop (##cdr l) (##fixnum.+ n 1))
      (if (##null? l) n (##fixnum.- 0 n)))))

(define ##interaction-cte
  (##make-top-cte))

;------------------------------------------------------------------------------

; Utilities

(define (##self-eval? val)
  (force-vars (val)
    (or (##complex? val)
        (##string? val)
        (##char? val)
        (##keyword? val)
        (##memq val '(#f #t #!eof #!optional #!rest #!key #!)))))

(define (##variable src x)
  (if (##not (##symbol? x))
    (ct-error-syntax "Identifier expected:" x))
  (if (##memq x
              '(quote quasiquote unquote unquote-splicing lambda if set!
                cond => else and or case let let* letrec begin do define
                delay future ##declare ##define-macro ##include
                c-define-type c-declare c-initialize c-lambda c-define))
    (ct-error-syntax "Variable name can not be a reserved identifier:" x)))

(define (##shape src x size)
  (let ((n (##proper-length x)))
    (if (or (##not n)
            (if (##fixnum.< 0 size)
              (##not (##fixnum.= n size))
              (##fixnum.< n (##fixnum.- 0 size))))
      (ct-error-syntax "Ill-formed special form:" (##car src)))))

(define (##proper-length l)

  (define (len l n)
    (cond ((##pair? l) (len (##cdr l) (##fixnum.+ n 1)))
          ((##null? l) n)
          (else        #f)))

  (len l 0))

(define (##force-list l)
  (if-forces
    (let loop ((l l))
      (force-vars (l)
        (if (##pair? l)
          (##cons (##car l) (loop (##cdr l)))
          l)))
    l))

(define (##read-expressions src filename)
  (if (##string? filename)

    (let ((port (##open-input-file filename)))
      (if port
        (##read-expressions-from-port
          port
          (lambda (script? exprs)
            exprs))
        (ct-error-syntax "File not found")))

    (ct-error-syntax "Filename expected")))

;------------------------------------------------------------------------------

; Compiler's main entry

(define (##compile-top top-cte src)
  (let ((tail? #f))
    (##comp-top top-cte src tail?)))

(define (##compile-inner cte src)
  (let ((tail? #f))
    (gen ##gen-top
      (##comp (##cte-frame cte (##list (self-var))) src tail?))))

;------------------------------------------------------------------------------

(define (##comp-top top-cte src tail?)
  (let ((src (##force-list src))
        (cte (##cte-top-cte top-cte)))
    (if (##pair? src)
      (let* ((first (##car src))
             (def (##macro-lookup cte first)))
        (if def
          (##comp-top top-cte (##macro-expand cte src def) tail?)
          (case first
            ((begin)          (##comp-top-begin top-cte src tail?))
            ((define)         (##comp-top-define top-cte src tail?))
            ((##declare)      (##comp-top-declare top-cte src tail?))
            ((##define-macro) (##comp-top-define-macro top-cte src tail?))
            ((##include)      (##comp-top-include top-cte src tail?))
            (else             (##comp-aux cte src tail? first)))))
      (##comp-simple cte src tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-begin top-cte src tail?)
  (##shape src src -1)
  (##comp-top-seq top-cte src tail? (##cdr src)))

(define (##comp-top-seq top-cte src tail? seq)
  (if (##pair? seq)
    (##comp-top-seq-aux top-cte src tail? seq)
    (let ((cte (##cte-top-cte top-cte)))
      (gen ##gen-cst (##void)))))

(define (##comp-top-seq-aux top-cte src tail? seq)
  (let ((rest (##cdr seq)))
    (if (##pair? rest)
      (let ((cte (##cte-top-cte top-cte)))
        (gen ##gen-seq
          (##comp-top top-cte (##car seq) #f)
          (##comp-top-seq-aux top-cte src tail? rest)))
      (##comp-top top-cte (##car seq) tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-define top-cte src tail?)
  (let ((name (##definition-name src)))
    (##variable src name)
    (let* ((cte (##cte-top-cte top-cte))
           (ind (##var-lookup cte name src)))
      (gen ##gen-glo-def
        ind
        (##comp cte (##definition-value src) #f)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-declare top-cte src tail?)
  (##shape src src -1)
  (let ((cte (##cte-top-cte top-cte)))
    (##top-cte-add-declarations! top-cte src)
    (gen ##gen-cst (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-define-macro top-cte src tail?)
  (let ((cte (##cte-top-cte top-cte))
        (name (##definition-name src)))
    (##top-cte-add-macro! top-cte name (##macro-def (##definition-value src)))
    (gen ##gen-cst (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-top-include top-cte src tail?)
  (##shape src src 2)
  (##comp-top-seq top-cte src tail? (##read-expressions src (##cadr src))))

;------------------------------------------------------------------------------

(define (##comp cte src tail?)
  (let ((src (##force-list src)))
    (if (##pair? src)
      (let* ((first (##car src))
             (def (##macro-lookup cte first)))
        (if def
          (##comp cte (##macro-expand cte src def) tail?)
          (case first
            ((begin)          (##comp-begin cte src tail?))
            ((define)         (ct-error-syntax "Ill-placed 'define'"))
            ((##declare)      (ct-error-syntax "Ill-placed '##declare'"))
            ((##define-macro) (ct-error-syntax "Ill-placed '##define-macro'"))
            ((##include)      (ct-error-syntax "Ill-placed '##include'"))
            (else             (##comp-aux cte src tail? first)))))
      (##comp-simple cte src tail?))))

(define (##comp-simple cte src tail?)
  (cond ((##symbol? src)    (##comp-ref cte src tail?))
        ((##self-eval? src) (##comp-cst cte src tail?))
        (else               (ct-error-syntax "Ill-formed expression"))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-begin cte src tail?)
  (##shape src src -2)
  (##comp-seq cte src tail? (##cdr src)))

(define (##comp-seq cte src tail? seq)
  (if (##pair? seq)
    (##comp-seq-aux cte src tail? seq)
    (gen ##gen-cst (##void))))

(define (##comp-seq-aux cte src tail? seq)
  (let ((rest (##cdr seq)))
    (if (##pair? rest)
      (gen ##gen-seq
        (##comp cte (##car seq) #f)
        (##comp-seq-aux cte src tail? rest))
      (##comp cte (##car seq) tail?))))

;------------------------------------------------------------------------------

(define (##comp-aux cte src tail? first)
  (case first
    ((quote)            (##comp-quote cte src tail?))
    ((quasiquote)       (##comp-quasiquote cte src tail?))
    ((unquote)          (ct-error-syntax "Ill-placed 'unquote'"))
    ((unquote-splicing) (ct-error-syntax "Ill-placed 'unquote-splicing'"))
    ((set!)             (##comp-set! cte src tail?))
    ((lambda)           (##comp-lambda cte src tail?))
    ((if)               (##comp-if cte src tail?))
    ((cond)             (##comp-cond cte src tail?))
    ((and)              (##comp-and cte src tail?))
    ((or)               (##comp-or cte src tail?))
    ((case)             (##comp-case cte src tail?))
    ((let)              (##comp-let cte src tail?))
    ((let*)             (##comp-let* cte src tail?))
    ((letrec)           (##comp-letrec cte src tail?))
    ((do)               (##comp-do cte src tail?))
    ((delay)            (##comp-delay cte src tail?))
    ((future)           (##comp-future cte src tail?))
    ((c-define-type)    (ct-error-syntax "'c-define-type' not supported by interpreter"))
    ((c-declare)        (ct-error-syntax "'c-declare' not supported by interpreter"))
    ((c-initialize)     (ct-error-syntax "'c-initialize' not supported by interpreter"))
    ((c-lambda)         (ct-error-syntax "'c-lambda' not supported by interpreter"))
    ((c-define)         (ct-error-syntax "'c-define' not supported by interpreter"))
    (else               (##comp-app cte src tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-ref cte src tail?)
  (##variable src src)
  (let ((x (##var-lookup cte src src)))
    (if (loc-access? x)
      (let ((up (loc-access-up x))
            (over (loc-access-over x)))
        (gen ##gen-loc-ref up over))
      (gen ##gen-glo-ref x))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-cst cte src tail?)
  (gen ##gen-cst src))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-quote cte src tail?)
  (##shape src src 2)
  (gen ##gen-cst (##cadr src)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-quasiquote cte src tail?)
  (##comp-quasi cte src tail? (##force-list (##cadr src)) 1))

(define (##comp-quasi cte src tail? form level)
  (cond ((##eq? level 0)
         (##comp cte form tail?))
        ((##pair? form)
         (let ((x (##car form)))
           (force-vars (x)
             (case x
               ((quasiquote)
                (##comp-quasi-list cte src tail? form (##fixnum.+ level 1)))
               ((unquote)
                (if (##eq? level 1)
                  (begin
                    (##shape src form 2)
                    (##comp cte (##cadr form) tail?))
                  (##comp-quasi-list cte src tail? form (##fixnum.- level 1))))
               ((unquote-splicing)
                (if (##eq? level 1)
                  (ct-error-syntax "Ill-placed 'unquote-splicing'"))
                (##comp-quasi-list cte src tail? form (##fixnum.- level 1)))
               (else
                (##comp-quasi-list cte src tail? form level))))))
        ((##vector? form)
         (gen ##gen-quasi-list->vector
           (##comp-quasi-list cte src #f (##vector->list form) level)))
        (else
         (gen ##gen-cst form))))

(define (##comp-quasi-list cte src tail? l level)
  (if (##pair? l)
    (let ((first (##force-list (##car l))))
      (if (and (##eq? level 1) (##unquote-splicing? first))
        (begin
          (##shape src first 2)
          (if (##null? (##cdr l))
            (##comp cte (##cadr first) tail?)
            (gen ##gen-quasi-append
              (##comp cte (##cadr first) #f)
              (##comp-quasi cte src #f (##cdr l) 1))))
        (gen ##gen-quasi-cons
          (##comp-quasi cte src #f first level)
          (##comp-quasi cte src #f (##cdr l) level))))
    (##comp-quasi cte src tail? l level)))

(define (##unquote-splicing? x)
  (and (##pair? x)
       (let ((y (##car x)))
         (force-vars (y)
           (##eq? y 'unquote-splicing)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-set! cte src tail?)
  (##shape src src 3)
  (let ((var (##cadr src)))
    (force-vars (var)
      (begin
        (##variable src var)
        (let ((x (##var-lookup cte var src)))
          (if (loc-access? x)
            (let ((up (loc-access-up x))
                  (over (loc-access-over x)))
              (gen ##gen-loc-set up over (##comp cte (##caddr src) #f)))
            (gen ##gen-glo-set x (##comp cte (##caddr src) #f))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-lambda cte src tail?)
  (##shape src src -3)
  (##comp-lambda-aux cte src tail? (##force-list (##cadr src)) (##cddr src)))

(define (##comp-lambda-aux cte src tail? parms body)
  (let* ((all-parms
          (##extract-parameters src parms))
         (required-parameters
          (##vector-ref all-parms 0))
         (optional-parameters
          (##vector-ref all-parms 1))
         (rest-parameter
          (##vector-ref all-parms 2))
         (key-parameters
          (##vector-ref all-parms 3)))
    (let loop1 ((frame required-parameters)
                (lst (or optional-parameters '()))
                (rev-inits '()))
      (if (##pair? lst)
        (let ((x (##car lst))
              (new-cte (##cte-frame cte (##cons (self-var) frame))))
          (loop1 (##append frame (##list (##car x)))
                 (##cdr lst)
                 (##cons (##comp new-cte (##cdr x) #f)
                         rev-inits)))
        (let loop2 ((frame (if rest-parameter
                             (##append frame (##list rest-parameter))
                             frame))
                    (lst (or key-parameters '()))
                    (rev-inits rev-inits)
                    (rev-keys '()))
          (if (##pair? lst)
            (let ((x (##car lst))
                  (new-cte (##cte-frame cte (##cons (self-var) frame))))
              (loop2 (##append frame (##list (##car x)))
                     (##cdr lst)
                     (##cons (##comp new-cte (##cdr x) #f)
                             rev-inits)
                     (##cons (##string->keyword (##symbol->string (##car x)))
                             rev-keys)))
            (let* ((new-cte (##cte-frame cte (##cons (self-var) frame)))
                   (c (##comp-body new-cte src #t body)))
              (cond ((or optional-parameters key-parameters)
                     (gen ##gen-prc
                          frame
                          rest-parameter
                          (and key-parameters
                               (##list->vector (##reverse rev-keys)))
                          c
                          (##reverse rev-inits)))
                    (rest-parameter
                     (gen ##gen-prc-rest frame c))
                    (else
                     (gen ##gen-prc-req frame c))))))))))

(define (##extract-parameters src parms)

  (define (parm-expected-err)
    (ct-error-syntax "Parameter must be an identifier"))

  (define (parm-or-default-binding-expected-err)
    (ct-error-syntax "Parameter must be an identifier or default binding"))

  (define (duplicate-parm-err)
    (ct-error-syntax "Duplicate parameter in parameter list"))

  (define (dotted-rest-err)
    (ct-error-syntax "Ill-placed dotted rest parameter"))

  (define (rest-parm-expected-err)
    (ct-error-syntax "#!rest must be followed by a parameter"))

  (define (default-binding-err)
    (ct-error-syntax "Ill-formed default binding"))

  (define (optional-illegal-err)
    (ct-error-syntax "Ill-placed #!optional"))

  (define (rest-illegal-err)
    (ct-error-syntax "Ill-placed #!rest"))

  (define (key-illegal-err)
    (ct-error-syntax "Ill-placed #!key"))

  (define (key-expected-err)
    (ct-error-syntax "#!key expected after rest parameter"))

  (define (default-binding-illegal-err)
    (ct-error-syntax "Ill-placed default binding"))

  (let loop ((lst (##force-list parms))
             (rev-required-parms '())
             (rev-optional-parms #f)
             (rest-parm #f)
             (rev-key-parms #f)
             (state 1)) ; 1 = required parms or #!optional or #!rest or #!key
                        ; 2 = optional parms or #!rest or #!key
                        ; 3 = #!key
                        ; 4 = key parms

    (define (done rest-parm)
      (##vector (##reverse rev-required-parms)
                (and rev-optional-parms (##reverse rev-optional-parms))
                rest-parm
                (if (or (##not rev-key-parms)
                        (and (##null? rev-key-parms) (##not rest-parm)))
                  #f
                  (##reverse rev-key-parms))))

    (define (check-if-duplicate parm)
      (if (or (##memq parm rev-required-parms)
              (and rev-optional-parms (##assq parm rev-optional-parms))
              (and rest-parm (##eq? parm rest-parm))
              (and rev-key-parms (##assq parm rev-key-parms)))
        (duplicate-parm-err)))

    (cond ((##null? lst)
           (done rest-parm))
          ((##pair? lst)
           (let ((parm (##car lst)))
             (force-vars (parm)
               (cond ((##eq? #!optional parm)
                      (if (##fixnum.< 1 state)
                        (optional-illegal-err))
                      (loop (##cdr lst)
                            rev-required-parms
                            '()
                            #f
                            #f
                            2))
                     ((##eq? #!rest parm)
                      (if (##fixnum.< 2 state)
                        (rest-illegal-err))
                      (if (##pair? (##cdr lst))
                        (let ((parm (##cadr lst)))
                          (force-vars (parm)
                            (begin
                              (##variable src parm)
                              (check-if-duplicate parm)
                              (loop (##cddr lst)
                                    rev-required-parms
                                    rev-optional-parms
                                    parm
                                    #f
                                    3))))
                        (rest-parm-expected-err)))
                     ((##eq? #!key parm)
                      (if (##fixnum.< 3 state)
                        (key-illegal-err))
                      (loop (##cdr lst)
                            rev-required-parms
                            rev-optional-parms
                            rest-parm
                            '()
                            4))
                     ((##fixnum.= state 3)
                      (key-expected-err))
                     ((##symbol? parm)
                      (##variable src parm)
                      (check-if-duplicate parm)
                      (case state
                        ((1)
                         (loop (##cdr lst)
                               (##cons parm
                                       rev-required-parms)
                               #f
                               #f
                               #f
                               1))
                        ((2)
                         (loop (##cdr lst)
                               rev-required-parms
                               (##cons (##cons parm #f)
                                       rev-optional-parms)
                               #f
                               #f
                               2))
                        (else
                         (loop (##cdr lst)
                               rev-required-parms
                               rev-optional-parms
                               rest-parm
                               (##cons (##cons parm #f)
                                       rev-key-parms)
                               4))))
                     ((##pair? parm)
                      (if (##not (or (##fixnum.= state 2) (##fixnum.= state 4)))
                        (default-binding-illegal-err))
                      (let ((parm (##force-list parm)))
                        (let ((length (##proper-length parm)))
                          (if (##not (##eqv? length 2))
                            (default-binding-err)))
                        (let ((parm (##car parm))
                              (val (##cadr parm)))
                          (force-vars (parm val)
                            (begin
                              (##variable src parm)
                              (check-if-duplicate parm)
                              (case state
                                ((2)
                                 (loop (##cdr lst)
                                       rev-required-parms
                                       (##cons (##cons parm val)
                                               rev-optional-parms)
                                       #f
                                       #f
                                       2))
                                (else
                                 (loop (##cdr lst)
                                       rev-required-parms
                                       rev-optional-parms
                                       rest-parm
                                       (##cons (##cons parm val)
                                               rev-key-parms)
                                       4))))))))
                     (else
                      (if (##fixnum.< 1 state)
                        (parm-or-default-binding-expected-err)
                        (parm-expected-err)))))))
          (else
           (let ((parm lst))
             (force-vars (parm)
               (begin
                 (##variable src parm)
                 (if (or rev-optional-parms
                         rest-parm
                         rev-key-parms)
                   (dotted-rest-err))
                 (check-if-duplicate parm)
                 (done parm))))))))

(define (##comp-body cte src tail? body)

  (define (letrec-defines cte vars vals body)
    (if (##pair? body)

      (let ((src (##force-list (##car body))))
        (if (##not (##pair? src))
          (letrec-defines* cte vars vals body)
          (let ((first (##car src)))
            (force-vars (first)
              (let ((def (##macro-lookup cte first)))
                (if def
                  (letrec-defines cte
                                  vars
                                  vals
                                  (##cons (##macro-expand cte src def)
                                          (##cdr body)))
                  (case first
                    ((begin)
                     (letrec-defines cte
                                     vars
                                     vals
                                     (##append (##cdr src) (##cdr body))))
                    ((define)
                     (let ((x (##definition-name src)))
                       (##variable src x)
                       (if (##memq x vars)
                         (ct-error-syntax "Duplicate definition of a variable"))
                       (letrec-defines cte
                                       (##cons x vars)
                                       (##cons (##definition-value src) vals)
                                       (##cdr body))))
                    ((##declare)
                     (##shape src src -1)
                     (letrec-defines (##cte-add-declarations cte src)
                                     vars
                                     vals
                                     (##cdr body)))
                    ((##define-macro)
                     (let ((x (##definition-name src)))
                       (letrec-defines (##cte-macro
                                         cte
                                         x
                                         (##macro-def (##definition-value src)))
                                       vars
                                       vals
                                       (##cdr body))))
                    ((##include)
                     (##shape src src 2)
                     (letrec-defines cte
                                     vars
                                     vals
                                     (##append (##read-expressions src (##cadr src))
                                               (##cdr body))))
                    (else
                     (letrec-defines* cte vars vals body)))))))))

      (ct-error-syntax "Body must contain at least one evaluable expression")))

  (define (letrec-defines* cte vars vals body)
    (if (##null? vars)
      (##comp-seq cte src tail? body)
      (##comp-letrec-aux cte src tail? vars vals body)))

  (letrec-defines cte '() '() body))

(define (##definition-name src)
  (##shape src src -3)
  (let ((pattern (##cadr src)))
    (force-vars (pattern)
      (let ((name (if (##pair? pattern)
                    (let ((name (##car pattern)))
                      (force-vars (name)
                        name))
                    (begin
                      (##shape src src 3)
                      pattern))))
        (if (##not (##symbol? name))
          (ct-error-syntax "Defined variable must be an identifier"))
        name))))

(define (##definition-value src)
  (let ((pattern (##cadr src)))
    (force-vars (pattern)
      (if (##pair? pattern)
        (##cons 'lambda (##cons (##cdr pattern) (##cddr src)))
        (##caddr src)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-if cte src tail?)
  (##shape src src -3)
  (if (##pair? (##cdddr src))
    (begin
      (##shape src src 4)
      (gen ##gen-if3
        (##comp cte (##cadr src) #f)
        (##comp cte (##caddr src) tail?)
        (##comp cte (##cadddr src) tail?)))
    (begin
      (##shape src src 3)
      (gen ##gen-if2
        (##comp cte (##cadr src) #f)
        (##comp cte (##caddr src) tail?)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-cond cte src tail?)
  (##shape src src -2)
  (##comp-cond-aux cte src tail? (##cdr src)))

(define (##comp-cond-aux cte src tail? clauses)
  (if (##pair? clauses)
    (let ((clause (##force-list (##car clauses))))
      (##shape src clause -1)
      (let ((x (##car clause)))
        (force-vars (x)
          (cond ((##eq? x 'else)
                 (##shape src clause -2)
                 (if (##not (##null? (##cdr clauses)))
                   (ct-error-syntax "ELSE clause must be last"))
                 (##comp-seq cte src tail? (##cdr clause)))
                ((##not (##pair? (##cdr clause)))
                 (gen ##gen-cond-or
                    (##comp cte (##car clause) #f)
                    (##comp-cond-aux cte src tail? (##cdr clauses))))
                (else
                 (let ((y (##cadr clause)))
                   (force-vars (y)
                     (if (##eq? y '=>)
                       (begin
                         (##shape src clause -3)
                         (gen ##gen-cond-send
                           (##comp cte (##car clause) #f)
                           (##comp cte (##caddr clause) #f)
                           (##comp-cond-aux cte src tail? (##cdr clauses))))
                       (gen ##gen-cond-if
                         (##comp cte (##car clause) #f)
                         (##comp-seq cte src tail? (##cdr clause))
                         (##comp-cond-aux cte src tail? (##cdr clauses)))))))))))
    (gen ##gen-cst (##void))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-and cte src tail?)
  (let ((rest (##cdr src)))
    (if (##pair? rest)
      (##comp-and-aux cte src tail? rest)
      (gen ##gen-cst #t))))

(define (##comp-and-aux cte src tail? l)
  (let ((rest (##cdr l)))
    (if (##pair? rest)
      (gen ##gen-and
        (##comp cte (##car l) #f)
        (##comp-and-aux cte src tail? rest))
      (##comp cte (##car l) tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-or cte src tail?)
  (let ((rest (##cdr src)))
    (if (##pair? rest)
      (##comp-or-aux cte src tail? rest)
      (gen ##gen-cst #f))))

(define (##comp-or-aux cte src tail? l)
  (let ((rest (##cdr l)))
    (if (##pair? rest)
      (gen ##gen-or
        (##comp cte (##car l) #f)
        (##comp-or-aux cte src tail? rest))
      (##comp cte (##car l) tail?))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-case cte src tail?)
  (##shape src src -3)
  (gen ##gen-case
    (##comp cte (##cadr src) #f)
    (let ((cte (##cte-frame cte (##list (selector-var)))))
      (##comp-case-aux cte src tail? (##cddr src)))))

(define (##comp-case-aux cte src tail? clauses)
  (if (##pair? clauses)
    (let ((clause (##force-list (##car clauses))))
      (##shape src clause -2)
      (let ((first (##force-list (##car clause))))
        (if (##eq? first 'else)
          (begin
            (if (##not (##null? (##cdr clauses)))
              (ct-error-syntax "ELSE clause must be last"))
            (gen ##gen-case-else
              (##comp-seq cte src tail? (##cdr clause))))
          (gen ##gen-case-clause
            first
            (##comp-seq cte src tail? (##cdr clause))
            (##comp-case-aux cte src tail? (##cdr clauses))))))
    (gen ##gen-case-else
      (gen ##gen-cst (##void)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-let cte src tail?)
  (##shape src src -3)
  (let ((x (##force-list (##cadr src))))
    (cond ((##symbol? x)
           (##shape src src -4)
           (let ((bindings (##force-list (##caddr src))))
             (let* ((vars (##bindings->vars src bindings #t))
                    (vals (##bindings->vals bindings)))
               (gen ##gen-app
                 (let ((inner-cte (##cte-frame cte (##list x))))
                   (gen ##gen-letrec
                     (##list x)
                     (let ((cte inner-cte)
                           (tail? #f))
                       (##list (gen ##gen-prc-req
                               vars
                               (##comp-body (##cte-frame
                                              cte
                                              (##cons (self-var) vars))
                                            src
                                            #t
                                            (##cdddr src)))))
                     (let ((cte inner-cte)
                           (tail? #f))
                       (gen ##gen-loc-ref 0 1)))) ; fetch loop variable
                 (##comp-vals cte vals)))))
          ((##null? x)
           (##comp-body cte src tail? (##cddr src)))
          (else
           (let* ((bindings x)
                  (vars (##bindings->vars src bindings #t))
                  (vals (##bindings->vals bindings)))
             (let ((c
                    (##comp-body
                      (##cte-frame cte vars)
                      src
                      tail?
                      (##cddr src))))
               (gen ##gen-let
                 vars
                 (##comp-vals cte vals)
                 c)))))))

(define (##comp-vals cte l)
  (if (##pair? l)
    (##cons (##comp cte (##car l) #f) (##comp-vals cte (##cdr l)))
    '()))

(define (##bindings->vars src bindings check-duplicates?)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##shape src binding 2)
      (let ((x (##car binding)))
        (force-vars (x)
          (let ((rest (##bindings->vars src (##cdr bindings) check-duplicates?)))
            (##variable src x)
            (if (and check-duplicates? (##memq x rest))
              (ct-error-syntax "Duplicate variable in bindings"))
            (##cons x rest)))))
    (if (##null? bindings)
      '()
      (ct-error-syntax "Ill-terminated bindings"))))

(define (##bindings->vals bindings)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##cons (##cadr binding) (##bindings->vals (##cdr bindings))))
    '()))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-let* cte src tail?)
  (##shape src src -3)
  (let ((bindings (##cadr src)))
    (force-vars (bindings)
      (let* ((vars (##bindings->vars src bindings #f))
             (vals (##bindings->vals bindings)))
        (##comp-let*-aux cte src tail? vars vals (##cddr src))))))

(define (##comp-let*-aux cte src tail? vars vals body)
  (if (##pair? vars)
    (let ((frame (##list (##car vars))))
      (let ((inner-cte (##cte-frame cte frame)))
        (gen ##gen-let
          frame
          (##list (##comp cte (##car vals) #f))
          (##comp-let*-aux inner-cte src tail? (##cdr vars) (##cdr vals) body))))
    (##comp-body cte src tail? body)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-letrec cte src tail?)
  (##shape src src -3)
  (let ((bindings (##force-list (##cadr src))))
    (if (##null? bindings)
      (##comp-body cte src tail? (##cddr src))
      (let* ((vars (##bindings->vars src bindings #t))
             (vals (##bindings->vals bindings)))
        (##comp-letrec-aux cte src tail? vars vals (##cddr src))))))

(define (##comp-letrec-aux cte src tail? vars vals body)
  (if (##pair? vars)
    (let ((inner-cte (##cte-frame cte vars)))
      (gen ##gen-letrec
        vars
        (##comp-vals inner-cte vals)
        (##comp-body inner-cte src tail? body)))
    (##comp-body cte src tail? body)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-do cte src tail?)
  (##shape src src -3)
  (let ((bindings (##force-list (##cadr src)))
        (exit (##force-list (##caddr src))))
    (##shape src exit -1)
    (let* ((vars (##bindings->vars* src bindings))
           (do-loop-vars (##list (do-loop-var)))
           (inner-cte (##cte-frame cte do-loop-vars)))
      (gen ##gen-letrec
        do-loop-vars
        (##list
          (let ((cte inner-cte)
                (tail? #f))
            (gen ##gen-prc-req
              vars
              (let ((cte (##cte-frame cte (##cons (self-var) vars)))
                    (tail? #t))
                (gen ##gen-if3
                  (##comp cte (##car exit) #f)
                  (##comp-seq cte src tail? (##cdr exit))
                  (let ((call
                          (gen ##gen-app
                            (let ((tail? #f))
                              (gen ##gen-loc-ref 1 1)) ; fetch do-loop-var
                            (##comp-vals cte (##bindings->steps bindings)))))
                    (if (##null? (##cdddr src))
                      call
                      (gen ##gen-seq
                        (##comp-seq cte src #f (##cdddr src))
                        call))))))))
        (let ((cte inner-cte))
          (gen ##gen-app
            (let ((tail? #f))
              (gen ##gen-loc-ref 0 1)) ; fetch do-loop-var
            (##comp-vals cte (##bindings->vals bindings))))))))

(define (##bindings->vars* src bindings)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##shape src binding -2)
      (if (##pair? (##cddr binding)) (##shape src binding 3))
      (let ((x (##car binding)))
        (force-vars (x)
          (let ((rest (##bindings->vars* src (##cdr bindings))))
            (##variable src x)
            (if (##memq x rest)
              (ct-error-syntax "Duplicate variable in bindings"))
            (##cons x rest)))))
    (if (##null? bindings)
      '()
      (ct-error-syntax "Ill-terminated bindings"))))

(define (##bindings->steps bindings)
  (if (##pair? bindings)
    (let ((binding (##force-list (##car bindings))))
      (##cons (if (##pair? (##cddr binding)) (##caddr binding) (##car binding))
              (##bindings->steps (##cdr bindings))))
    '()))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-app cte src tail?)
  (let ((n (##proper-length src)))
    (if n
      (gen ##gen-app
        (##comp cte (##car src) #f)
        (##comp-vals cte (##cdr src)))
      (ct-error-syntax "Ill-formed procedure application"))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-delay cte src tail?)
  (##shape src src 2)
  (gen ##gen-delay (##comp cte (##cadr src) #t)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##comp-future cte src tail?)
  (##shape src src 2)
  (gen ##gen-future (##comp cte (##cadr src) #t)))

;==============================================================================

; Code generation procedures

;------------------------------------------------------------------------------

(define ##cprc-top
  (mk-cprc
    (let ((rte (mk-rte rte #f)))
      (code-run (^ 0)))))

(define ##gen-top
  (mk-gen (val)
    (mk-code ##cprc-top cte (val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-cst-null  (mk-cprc '()))
(define ##cprc-cst-true  (mk-cprc #t))
(define ##cprc-cst-false (mk-cprc #f))
(define ##cprc-cst--2    (mk-cprc -2))
(define ##cprc-cst--1    (mk-cprc -1))
(define ##cprc-cst-0     (mk-cprc 0))
(define ##cprc-cst-1     (mk-cprc 1))
(define ##cprc-cst-2     (mk-cprc 2))
(define ##cprc-cst       (mk-cprc (^ 0)))

(define ##gen-cst
  (mk-gen (val)
    (case val
      ((()) (mk-code ##cprc-cst-null  cte ()))
      ((#t) (mk-code ##cprc-cst-true  cte ()))
      ((#f) (mk-code ##cprc-cst-false cte ()))
      ((-2) (mk-code ##cprc-cst--2    cte ()))
      ((-1) (mk-code ##cprc-cst--1    cte ()))
      ((0)  (mk-code ##cprc-cst-0     cte ()))
      ((1)  (mk-code ##cprc-cst-1     cte ()))
      ((2)  (mk-code ##cprc-cst-2     cte ()))
      (else (mk-code ##cprc-cst       cte () val)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-loc-ref-0-1 (mk-cprc (rte-ref rte 1)))
(define ##cprc-loc-ref-0-2 (mk-cprc (rte-ref rte 2)))
(define ##cprc-loc-ref-0-3 (mk-cprc (rte-ref rte 3)))

(define ##cprc-loc-ref-1-1 (mk-cprc (rte-ref (rte-up rte) 1)))
(define ##cprc-loc-ref-1-2 (mk-cprc (rte-ref (rte-up rte) 2)))
(define ##cprc-loc-ref-1-3 (mk-cprc (rte-ref (rte-up rte) 3)))

(define ##cprc-loc-ref-2-1 (mk-cprc (rte-ref (rte-up (rte-up rte)) 1)))
(define ##cprc-loc-ref-2-2 (mk-cprc (rte-ref (rte-up (rte-up rte)) 2)))
(define ##cprc-loc-ref-2-3 (mk-cprc (rte-ref (rte-up (rte-up rte)) 3)))

(define ##cprc-loc-ref
  (mk-cprc
    (let loop ((e rte) (i (^ 0)))
      (if (##fixnum.< 0 i)
        (loop (rte-up e) (##fixnum.- i 1))
        (rte-ref e (^ 1))))))

(define ##gen-loc-ref
  (mk-gen (up over)
    (case up
      ((0)
       (case over
         ((1)  (mk-code ##cprc-loc-ref-0-1 cte ()))
         ((2)  (mk-code ##cprc-loc-ref-0-2 cte ()))
         ((3)  (mk-code ##cprc-loc-ref-0-3 cte ()))
         (else (mk-code ##cprc-loc-ref     cte () up over))))
      ((1)
       (case over
         ((1)  (mk-code ##cprc-loc-ref-1-1 cte ()))
         ((2)  (mk-code ##cprc-loc-ref-1-2 cte ()))
         ((3)  (mk-code ##cprc-loc-ref-1-3 cte ()))
         (else (mk-code ##cprc-loc-ref     cte () up over))))
      ((2)
       (case over
         ((1)  (mk-code ##cprc-loc-ref-2-1 cte ()))
         ((2)  (mk-code ##cprc-loc-ref-2-2 cte ()))
         ((3)  (mk-code ##cprc-loc-ref-2-3 cte ()))
         (else (mk-code ##cprc-loc-ref     cte () up over))))
     (else
       (mk-code ##cprc-loc-ref cte () up over)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-glo-ref
  (mk-cprc
    (let loop ((val (global-var-ref (^ 0))))
      (if (unbound? val)
        (loop (rt-error-unbound-global-var $code rte))
        val))))

(define ##gen-glo-ref
  (mk-gen (ind)
    (mk-code ##cprc-glo-ref cte () ind)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-loc-set
  (mk-cprc
    (let ((val (code-run (^ 0))))
      (let loop ((e rte) (i (^ 1)))
        (if (##fixnum.< 0 i)
          (loop (rte-up e) (##fixnum.- i 1))
          (begin
            (rte-set! e (^ 2) val)
            (##void)))))))

(define ##gen-loc-set
  (mk-gen (up over val)
    (mk-code ##cprc-loc-set cte (val) up over)))

(define ##cprc-glo-set
  (mk-cprc
    (let ((val (code-run (^ 0))))
      (global-var-set! (^ 1) val)
      (##void))))

(define ##gen-glo-set
  (mk-gen (ind val)
    (mk-code ##cprc-glo-set cte (val) ind)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-glo-def
  (mk-cprc
    (let ((rte #f))
      (global-var-set! (^ 1) (code-run (^ 0)))
      (##void))))

(define ##gen-glo-def
  (mk-gen (ind val)
    (mk-code ##cprc-glo-def cte (val) ind)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-if2
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (code-run (^ 1))
          (##void))))))

(define ##gen-if2
  (mk-gen (pre con)
    (mk-code ##cprc-if2 cte (pre con))))

(define ##cprc-if3
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (code-run (^ 1))
          (code-run (^ 2)))))))

(define ##gen-if3
  (mk-gen (pre con alt)
    (mk-code ##cprc-if3 cte (pre con alt))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-seq
  (mk-cprc
    (code-run (^ 0))
    (code-run (^ 1))))

(define ##gen-seq
  (mk-gen (val1 val2)
    (mk-code ##cprc-seq cte (val1 val2))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-quasi-list->vector
  (mk-cprc
    (quasi-list->vector (code-run (^ 0)))))

(define ##gen-quasi-list->vector
  (mk-gen (val)
    (mk-code ##cprc-quasi-list->vector cte (val))))

(define ##cprc-quasi-append
  (mk-cprc
    (quasi-append (code-run (^ 0)) (code-run (^ 1)))))

(define ##gen-quasi-append
  (mk-gen (val1 val2)
    (mk-code ##cprc-quasi-append cte (val1 val2))))

(define ##cprc-quasi-cons
  (mk-cprc
    (quasi-cons (code-run (^ 0)) (code-run (^ 1)))))

(define ##gen-quasi-cons
  (mk-gen (val1 val2)
    (mk-code ##cprc-quasi-cons cte (val1 val2))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-cond-if
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (code-run (^ 1))
          (code-run (^ 2)))))))

(define ##gen-cond-if
  (mk-gen (val1 val2 val3)
    (mk-code ##cprc-cond-if cte (val1 val2 val3))))

(define ##cprc-cond-or
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          pred
          (code-run (^ 1)))))))

(define ##gen-cond-or
  (mk-gen (val1 val2)
    (mk-code ##cprc-cond-or cte (val1 val2))))

(define ##cprc-cond-send-red
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (let loop ((proc (code-run (^ 1))))
            (force-vars (proc)
              (if (##not (##procedure? proc))
                (loop (rt-error-non-procedure-send $code rte))
                (##reduction-apply1 $code rte proc pred))))
          (code-run (^ 2)))))))

(define ##cprc-cond-send-sub
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          (let loop ((proc (code-run (^ 1))))
            (force-vars (proc)
              (if (##not (##procedure? proc))
                (loop (rt-error-non-procedure-send $code rte))
                (##subproblem-apply1 $code rte proc pred))))
          (code-run (^ 2)))))))

(define ##gen-cond-send
  (mk-gen (val1 val2 val3)
    (mk-code (if tail? ##cprc-cond-send-red ##cprc-cond-send-sub)
             cte
             (val1 val2 val3))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-or
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (true? pred)
          pred
          (code-run (^ 1)))))))

(define ##gen-or
  (mk-gen (val1 val2)
    (mk-code ##cprc-or cte (val1 val2))))

(define ##cprc-and
  (mk-cprc
    (let ((pred (code-run (^ 0))))
      (force-vars (pred)
        (if (##not (true? pred))
          pred
          (code-run (^ 1)))))))

(define ##gen-and
  (mk-gen (val1 val2)
    (mk-code ##cprc-and cte (val1 val2))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-case
  (mk-cprc
    (let ((selector (code-run (^ 0))))
      (force-vars (selector)
        (let ((rte (mk-rte rte selector)))
          (code-run (^ 1)))))))

(define ##gen-case
  (mk-gen (val1 val2)
    (mk-code ##cprc-case cte (val1 val2))))

(define ##cprc-case-clause
  (mk-cprc
    (if (##case-memv (rte-ref rte 1) (^ 2))
      (code-run (^ 0))
      (code-run (^ 1)))))

(define ##gen-case-clause
  (mk-gen (cases val1 val2)
    (mk-code ##cprc-case-clause cte (val1 val2) cases)))

(define ##cprc-case-else
  (mk-cprc
    (code-run (^ 0))))

(define ##gen-case-else
  (mk-gen (val)
    (mk-code ##cprc-case-else cte (val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-let
  (mk-cprc
    (let ((ns (##fixnum.- (code-length $code) 2)))
      (let loop1 ((i 1) (args '()))
        (if (##fixnum.< ns i)
          (let ((inner-rte (mk-rte* rte ns)))
            (let loop2 ((i ns) (args args))
              (if (##fixnum.< 0 i)
                (begin
                  (rte-set! inner-rte i (##car args))
                  (loop2 (##fixnum.- i 1) (##cdr args)))
                (let ((rte inner-rte))
                  (code-run (^ 0))))))
          (loop1 (##fixnum.+ i 1)
                 (##cons (code-run (code-ref $code i)) args)))))))

(define ##gen-let
  (mk-gen (vars vals body)
    (let ((c (##mk-code* ##cprc-let cte (##cons body vals) 1)))
      (code-set! c (##fixnum.+ (##length vals) 1) vars)
      c)))

(define ##cprc-letrec
  (mk-cprc
    (let ((ns (##fixnum.- (code-length $code) 2)))
      (let ((rte (mk-rte* rte ns)))
        (let loop ((i ns))
          (if (##fixnum.< 0 i)
            (begin
              (rte-set! rte i (code-run (code-ref $code i)))
              (loop (##fixnum.- i 1)))
            (code-run (^ 0))))))))

(define ##gen-letrec
  (mk-gen (vars vals body)
    (let ((c (##mk-code* ##cprc-letrec cte (##cons body vals) 1)))
      (code-set! c (##fixnum.+ (##length vals) 1) vars)
      c)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-prc-req0
  (mk-cprc
    (letrec ((proc
               (lambda ()
                 (let ((rte (mk-rte rte proc)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc-req1
  (mk-cprc
    (letrec ((proc
               (lambda (arg1)
                 (let ((rte (mk-rte rte proc arg1)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc-req2
  (mk-cprc
    (letrec ((proc
               (lambda (arg1 arg2)
                 (let ((rte (mk-rte rte proc arg1 arg2)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc-req3
  (mk-cprc
    (letrec ((proc
               (lambda (arg1 arg2 arg3)
                 (let ((rte (mk-rte rte proc arg1 arg2 arg3)))
                   (code-run (^ 0))))))
      proc)))

(define ##cprc-prc-req
  (mk-cprc
    (letrec ((proc
               (lambda args
                 (let ((ns (^ 1)))
                   (let ((inner-rte (mk-rte* rte ns)))
                     (rte-set! inner-rte 1 proc)
                     (let loop ((i 2) (l args))
                       (if (##fixnum.< ns i)
                         (if (##pair? l)
                           (rt-error-wrong-nb-args proc args)
                           (let ((rte inner-rte))
                             (code-run (^ 0))))
                         (if (##pair? l)
                           (begin
                             (rte-set! inner-rte i (##car l))
                             (loop (##fixnum.+ i 1) (##cdr l)))
                           (rt-error-wrong-nb-args proc args)))))))))
      proc)))

(define ##gen-prc-req
  (mk-gen (frame body)
    (let ((n (##length frame)))
      (case n
        ((0)  (mk-code ##cprc-prc-req0 cte (body) frame))
        ((1)  (mk-code ##cprc-prc-req1 cte (body) frame))
        ((2)  (mk-code ##cprc-prc-req2 cte (body) frame))
        ((3)  (mk-code ##cprc-prc-req3 cte (body) frame))
        (else (mk-code ##cprc-prc-req  cte (body) (##fixnum.+ n 1) frame))))))

(define ##cprc-prc-rest
  (mk-cprc
    (letrec ((proc
               (lambda args
                 (let ((ns (^ 1)))
                   (let ((inner-rte (mk-rte* rte ns)))
                     (rte-set! inner-rte 1 proc)
                     (let loop ((i 2) (l args))
                       (if (##fixnum.< i ns)
                         (if (##pair? l)
                           (begin
                             (rte-set! inner-rte i (##car l))
                             (loop (##fixnum.+ i 1) (##cdr l)))
                           (rt-error-wrong-nb-args proc args))
                         (begin
                           (rte-set! inner-rte i l)
                           (let ((rte inner-rte))
                             (code-run (^ 0)))))))))))
      proc)))

(define ##gen-prc-rest
  (mk-gen (frame body)
    (let ((n (##length frame)))
      (mk-code ##cprc-prc-rest cte (body) (##fixnum.+ n 1) frame))))

(define ##cprc-prc
  (mk-cprc
    (letrec ((proc
               (lambda args
                 (let* ((n
                         (code-length $code))
                        (inner-rte
                         (mk-rte* rte (code-ref $code (##fixnum.- n 6)))))

                   (define (get-keys i j l rest? keys)
                     (let loop1 ((lst l))
                       (if (##pair? lst)
                         (let ((key (##car lst)) (lst (##cdr lst)))
                           (if (##keyword? key)
                             (if rest?
                               (if (##pair? lst)
                                 (loop1 (##cdr lst))
                                 (rt-error-wrong-nb-args proc args))
                               (let loop2 ((k (##fixnum.-
                                               (##vector-length keys)
                                               1)))
                                 (cond ((##fixnum.< k 0)
                                        (rt-error-unknown-keyword-arg proc args))
                                       ((##eq? key (##vector-ref keys k))
                                        (if (##pair? lst)
                                          (loop1 (##cdr lst))
                                          (rt-error-wrong-nb-args proc args)))
                                       (else
                                        (loop2 (##fixnum.- k 1))))))
                             (rt-error-keyword-expected proc args)))
                         (let loop3 ((i i) (j j) (k 0))
                           (if (##fixnum.< k (##vector-length keys))
                             (let ((key (##vector-ref keys k)))
                               (let loop4 ((lst l))
                                 (if (##pair? lst)
                                   (if (##eq? (##car lst) key)
                                     (begin
                                       (rte-set! inner-rte i (##cadr lst))
                                       (loop3 (##fixnum.+ i 1)
                                              (##fixnum.+ j 1)
                                              (##fixnum.+ k 1)))
                                     (loop4 (##cddr lst)))
                                   (let ((rte inner-rte))
                                     (rte-set! inner-rte i
                                       (code-run (code-ref $code j)))
                                     (loop3 (##fixnum.+ i 1)
                                            (##fixnum.+ j 1)
                                            (##fixnum.+ k 1))))))
                             (let ((rte inner-rte))
                               (code-run (^ 0))))))))

                   (rte-set! inner-rte 1 proc)
                   (let loop1 ((i 2) (l args))
                     (if (##fixnum.< i (code-ref $code (##fixnum.- n 5)))
                       (if (##pair? l)
                         (begin
                           (rte-set! inner-rte i (##car l))
                           (loop1 (##fixnum.+ i 1) (##cdr l)))
                         (rt-error-wrong-nb-args proc args))
                       (let loop2 ((i i) (j 1) (l l))
                         (if (##fixnum.< i (code-ref $code (##fixnum.- n 4)))
                           (if (##pair? l)
                             (begin
                               (rte-set! inner-rte i (##car l))
                               (loop2 (##fixnum.+ i 1)
                                      (##fixnum.+ j 1)
                                      (##cdr l)))
                             (let ((rte inner-rte))
                               (rte-set! inner-rte i
                                 (code-run (code-ref $code j)))
                               (loop2 (##fixnum.+ i 1)
                                      (##fixnum.+ j 1)
                                      '())))
                           (let ((keys (code-ref $code (##fixnum.- n 2))))
                             (cond ((code-ref $code (##fixnum.- n 3))
                                    (rte-set! inner-rte i l)
                                    (if keys
                                      (get-keys (##fixnum.+ i 1) j l #t keys)
                                      (let ((rte inner-rte))
                                        (code-run (^ 0)))))
                                   (keys
                                    (get-keys i j l #f keys))
                                   ((##null? l)
                                    (let ((rte inner-rte))
                                      (code-run (^ 0))))
                                   (else
                                    (rt-error-wrong-nb-args proc args))))))))))))
      proc)))

(define ##gen-prc
  (mk-gen (frame rest? keys body inits)
    (let* ((n (##length frame))
           (ni (##length inits))
           (nr (##fixnum.- (##fixnum.- n ni) (if rest? 1 0)))
           (no (##fixnum.- ni (if keys (##vector-length keys) 0)))
           (c (##mk-code* ##cprc-prc cte (##cons body inits) 6)))
      (code-set! c (##fixnum.+ ni 1) (##fixnum.+ n 1))
      (code-set! c (##fixnum.+ ni 2) (##fixnum.+ nr 2))
      (code-set! c (##fixnum.+ ni 3) (##fixnum.+ (##fixnum.+ nr 2) no))
      (code-set! c (##fixnum.+ ni 4) rest?)
      (code-set! c (##fixnum.+ ni 5) keys)
      (code-set! c (##fixnum.+ ni 6) frame)
      c)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-app0-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (##reduction-apply0 $code rte proc))))))

(define ##cprc-app1-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1))))
            (##reduction-apply1 $code rte proc arg1)))))))

(define ##cprc-app2-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2))))
            (##reduction-apply2 $code rte proc arg1 arg2)))))))

(define ##cprc-app3-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2)))
                (arg3 (code-run (^ 3))))
            (##reduction-apply3 $code rte proc arg1 arg2 arg3)))))))

(define ##cprc-app-red
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let loop ((i (##fixnum.- (code-length $code) 1)) (args '()))
            (if (##fixnum.< 0 i)
              (loop (##fixnum.- i 1)
                    (##cons (code-run (code-ref $code i)) args))
              (##reduction-apply $code rte proc args))))))))

(define ##cprc-app0-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (##subproblem-apply0 $code rte proc))))))

(define ##cprc-app1-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1))))
            (##subproblem-apply1 $code rte proc arg1)))))))

(define ##cprc-app2-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2))))
            (##subproblem-apply2 $code rte proc arg1 arg2)))))))

(define ##cprc-app3-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let ((arg1 (code-run (^ 1)))
                (arg2 (code-run (^ 2)))
                (arg3 (code-run (^ 3))))
            (##subproblem-apply3 $code rte proc arg1 arg2 arg3)))))))

(define ##cprc-app-sub
  (mk-cprc
    (let ((proc (code-run (^ 0))))
      (force-vars (proc)
        (if (##not (##procedure? proc))
          (rt-error-non-procedure-oper $code rte)
          (let loop ((i (##fixnum.- (code-length $code) 1)) (args '()))
            (if (##fixnum.< 0 i)
              (loop (##fixnum.- i 1)
                    (##cons (code-run (code-ref $code i)) args))
              (##subproblem-apply $code rte proc args))))))))

(define ##properly-tail-recursive? #f)
(set! ##properly-tail-recursive? #t)

(define ##gen-app
  (mk-gen (oper args)
    (if (and tail? ##properly-tail-recursive?)
      (case (##length args)
        ((0)
         (mk-code ##cprc-app0-red
                  cte
                  (oper)))
        ((1)
         (mk-code ##cprc-app1-red
                  cte
                  (oper (##car args))))
        ((2)
         (mk-code ##cprc-app2-red
                  cte
                  (oper (##car args) (##cadr args))))
        ((3)
         (mk-code ##cprc-app3-red
                  cte
                  (oper (##car args) (##cadr args) (##caddr args))))
        (else
         (##mk-code* ##cprc-app-red
                     cte
                     (##cons oper args)
                     0)))
      (case (##length args)
        ((0)
         (mk-code ##cprc-app0-sub
                  cte
                  (oper)))
        ((1)
         (mk-code ##cprc-app1-sub
                  cte
                  (oper (##car args))))
        ((2)
         (mk-code ##cprc-app2-sub
                  cte
                  (oper (##car args) (##cadr args))))
        ((3)
         (mk-code ##cprc-app3-sub
                  cte
                  (oper (##car args) (##cadr args) (##caddr args))))
        (else
         (##mk-code* ##cprc-app-sub
                     cte
                     (##cons oper args)
                     0))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##reduction-apply0 $code rte proc)
  (##declare (interrupts-enabled))
  (proc))

(define (##reduction-apply1 $code rte proc arg1)
  (##declare (interrupts-enabled))
  (proc arg1))

(define (##reduction-apply2 $code rte proc arg1 arg2)
  (##declare (interrupts-enabled))
  (proc arg1 arg2))

(define (##reduction-apply3 $code rte proc arg1 arg2 arg3)
  (##declare (interrupts-enabled))
  (proc arg1 arg2 arg3))

(define (##reduction-apply $code rte proc args)
  (##declare (interrupts-enabled))
  (##apply proc args))

(define ##subproblem-apply0
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc))
                  $code
                  rte))))

(define ##subproblem-apply1
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc arg1)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc arg1))
                  $code
                  rte))))

(define ##subproblem-apply2
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc arg1 arg2)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc arg1 arg2))
                  $code
                  rte))))

(define ##subproblem-apply3
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc arg1 arg2 arg3)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (proc arg1 arg2 arg3))
                  $code
                  rte))))

(define ##subproblem-apply
  (let ()
    (##declare (not inline))
    (lambda ($code rte proc args)
      (##declare (not interrupts-enabled))
      (##identity (let ()
                    (##declare (interrupts-enabled))
                    (##apply proc args))
                  $code
                  rte))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-delay
  (mk-cprc
    (delay (code-run (^ 0)))))

(define ##gen-delay
  (mk-gen (val)
    (mk-code ##cprc-delay cte (val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##cprc-future
  (mk-cprc
    (future (code-run (^ 0)))))

(define ##gen-future
  (mk-gen (val)
    (mk-code ##cprc-future cte (val))))

;==============================================================================

; Eval

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Evaluation in a top environment with the current dynamic environment.

(define ##eval-top #f)
(set! ##eval-top
  (lambda (expr top-cte)
    (##eval-top-with-dyn-env expr top-cte (##dynamic-env-ref))))

; Evaluation in a top environment with a specific dynamic environment.

(define ##eval-top-with-dyn-env #f)
(set! ##eval-top-with-dyn-env
  (lambda (expr top-cte dyn-env)
    (let ((c (##compile-top top-cte expr)))
      (##dynamic-env-bind
        dyn-env
        (lambda () (let ((rte #f)) (code-run c)))))))

; Evaluation inside a specific environment ('cte' describes the runtime
; environment 'rte').

(define ##eval-inner #f)
(set! ##eval-inner
  (lambda (expr cte rte dyn-env)
    (let ((c (##compile-inner cte expr)))
      (##dynamic-env-bind
        dyn-env
        (lambda () (let ((rte rte)) (code-run c)))))))

(define (eval expr #!optional env)
  (##eval-top expr ##interaction-cte))

;==============================================================================

(define (##read-expressions-from-port port cont)

  (define (read-exprs)
    (let ((expr (##read port)))
      (if (##not (##eof-object? expr))
        (##cons expr (read-exprs))
        '())))

  (##cleanup-on-signal
    (lambda ()
      (let ((first (##read port)))
        (let* ((script?
                (and (##eq? first '#!)
                     (let loop ()
                       (let ((c (##read-char port)))
                         (if (or (##eof-object? c) (##char=? c #\newline))
                           #t
                           (loop))))))
               (rest
                (read-exprs)))
          (##close-port port)
          (cond ((##eof-object? first)
                 (cont #f '()))
                (script?
                 (cont #t rest))
                (else
                 (cont #f (##cons first rest)))))))
    (lambda ()
      (##close-port port))))

(define (##load name type-callback trap-if-not-found? char-encoding)

  (define (load-source-from-port port)
    (##read-expressions-from-port
      port
      (lambda (script? exprs)
        (type-callback script?)
        (##eval-top (##cons 'begin exprs)
                    (##top-cte-clone ##interaction-cte)))))

  (define (load-source name name-with-ext)
    (let ((port (##open-input-file name-with-ext)))
      (if port
        (begin
          (load-source-from-port port)
          (##port-name port))
        (if trap-if-not-found?
          (trap-open-file (load name))
          #f))))

  (define (load-binary name path)
    (let ((msg (##load-object-file path)))
      (if (##procedure? msg)
        (begin
          (msg)
          path)
        (trap-load (load name) msg))))
 
  (define (load-no-ext name)
    (let ((port (##open-input-file name)))
      (if port
        (begin
          (load-source-from-port port)
          (##port-name port))
        (let loop ((version 1) (last #f))
          (let ((obj-port
                 (##open-input-file
                   (##string-append name
                                    ".o"
                                    (##number->string version 10)))))
            (if obj-port
              (begin
                (##close-port obj-port)
                (loop (##fixnum.+ version 1) (##port-name obj-port)))
              (if last
                (load-binary name last)
                (load-source name (##string-append name ".scm")))))))))

  (define (binary-extension? ext)
    (let ((len (##string-length ext)))
      (and (##fixnum.< 2 len)
           (##char=? (##string-ref ext 0) #\.)
           (##char=? (##string-ref ext 1) #\o)
           (let ((c (##string-ref ext 2)))
             (and (##char>=? c #\1) (##char<=? c #\9)
                  (let loop ((i (##fixnum.- len 1)))
                    (if (##fixnum.< i 3)
                      #t
                      (let ((c (##string-ref ext i)))
                        (and (##char>=? c #\0) (##char<=? c #\9)
                             (loop (##fixnum.- i 1)))))))))))

  (let ((ext (##path-extension name)))
    (cond ((##string=? ext "")
           (load-no-ext name))
          ((binary-extension? ext)
           (let ((path (##path-expand name)))
             (if path
               (load-binary name path)
               (if trap-if-not-found?
                 (trap-open-file (load name))
                 #f))))
          (else
           (load-source name name)))))

(define (load path #!optional (char-encoding (absent-obj)))
  (force-vars (path char-encoding)
    (check-string path (load path char-encoding)
      (##load
       path
       (lambda (script?) #f)
       #t
       (if (##eq? char-encoding (absent-obj)) #f char-encoding)))))

;==============================================================================
